#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import pandas as pd
import plotly.express as px
import plotly.graph_objects as go


def _get_fairness_details(run_details, pipeline_name):
    """
    Returns fairness details based on experiment run details.
    @param run_details: dict
    @param pipeline_name: str
    @return: dict
    """
    for pipeline in run_details['entity']['status'].get('metrics', []):
        if pipeline['context']['intermediate_model']['name'].split('P')[-1] == pipeline_name.split('_')[-1]:
            return pipeline['context']['fairness']


def get_pipeline_fairness_score(fairness_details):
    return fairness_details['visualization_data']['holdout']['monitored_group']['fairness']['value']


def get_favorable_outcomes(fairness_details):
    return fairness_details['visualization_data']['holdout']['monitored_group']['favorable_outcome']['values']


def _get_pipeline_percentage_of(fairness_details, group, outcome):
    return fairness_details['visualization_data']['holdout'][group][outcome]['percentage']


def _get_pipeline_perfect_equality(fairness_details):
    return fairness_details['visualization_data']['holdout']['monitored_group']['favorable_outcome']['perfect_equality']['percentage']


def get_attributes_scores(fairness_details):
    return {attr['feature']: attr['fairness']['percentage'] for attr in fairness_details['visualization_data']['holdout']['monitored_group']['protected_attributes']}


def _get_attributes_perfect_equality(fairness_details):
    return {attr['feature']: attr['favorable_outcome']['perfect_equality']['percentage'] for attr in fairness_details['visualization_data']['holdout']['monitored_group']['protected_attributes']}


def _get_pipeline_fairness(fairness_details):
    """
    Returns pandas dataframe with fairness descriptive statics for pipeline model.
    @param fairness_details: dict
    @return: pandas df
    """

    favorable_outcomes_percentage_monitored = _get_pipeline_percentage_of(fairness_details, 'monitored_group', 'favorable_outcome')
    favorable_outcomes_percentage_reference = _get_pipeline_percentage_of(fairness_details, 'reference_group', 'favorable_outcome')

    data_model = [['favorable', 'monitored', favorable_outcomes_percentage_monitored],
                  ['favorable', 'reference', favorable_outcomes_percentage_reference],
                  ['unfavorable', 'monitored', 100 - favorable_outcomes_percentage_monitored],
                  ['unfavorable', 'reference', 100 - favorable_outcomes_percentage_reference]]

    return pd.DataFrame(data_model, columns=['Outcome', 'Group', '%'])


def get_protected_attributes(fairness_details):
    return [attr['feature'] for attr in fairness_details['visualization_data']['holdout']['monitored_group']['protected_attributes']]


def _get_attributes_fairness(fairness_details):
    attr_data = []

    for group in ['monitored_group', 'reference_group']:
        attributes = fairness_details['visualization_data']['holdout'][group]['protected_attributes']
        for attr in attributes:
            name = attr['feature']

            if len(attributes) > 1:
                fav_percentage = attr['favorable_outcome']['percentage']
            else:
                fav_percentage = _get_pipeline_percentage_of(fairness_details, group, 'favorable_outcome')

            attr_data.append(['unfavorable', group, name, 100 - fav_percentage])
            attr_data.append(['favorable', group, name, fav_percentage])

    return pd.DataFrame(attr_data, columns=['Outcome', 'Group', 'Feature', '%'])


def visualize(run_details: dict, pipeline_name: str):
    """
    Plots favorable outcome distributions using plotly package.
    @param run_details: dict
    @param pipeline_name: str
    """
    if not isinstance(run_details, dict):
        raise ValueError('run_details should be dict not {0}'.format(type(run_details)))

    if not isinstance(pipeline_name, str):
        raise ValueError('pipeline_name should be str not {0}'.format(type(pipeline_name)))

    colors = ["#2A66DE", "gray"]
    fairness_details = _get_fairness_details(run_details=run_details, pipeline_name=pipeline_name)
    df = _get_pipeline_fairness(fairness_details)
    pipeline_equality = _get_pipeline_perfect_equality(fairness_details)
    df_attr = _get_attributes_fairness(fairness_details)
    fig1 = px.bar(df, y="Group", x="%", color="Outcome", title="Pipeline", orientation='h', color_discrete_sequence=colors)
    fig1.add_vline(x=pipeline_equality, line_width=1, line_color="black", annotation_text="Perfect equality", annotation_position="top left")
    fig1.show()

    fig2 = go.Figure()
    fig2.update_layout(
        title="Protected attributes",
        template="simple_white",
        xaxis=dict(title_text="Outcome [%]"),
        yaxis=dict(title_text="Protected attribute"),
        barmode="stack"
    )

    for r, c in zip(sorted(df_attr.Outcome.unique()), colors):
        plot_df = df_attr[df_attr.Outcome == r]
        fig2.add_trace(go.Bar(y=[plot_df.Feature, plot_df.Group], x=plot_df['%'], name=r, marker_color=c, orientation='h'))
    fig2.show()



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest

from models_preparation import *
from preparation_and_cleaning import *
import logging


class TestWMLClientWithScikitLearn(unittest.TestCase):
    model_uid = None
    scoring_url = None
    logger = logging.getLogger(__name__)

    @classmethod
    def setUpClass(self):
        TestWMLClientWithScikitLearn.logger.info("Service Instance: setting up credentials")

        self.wml_credentials = get_wml_credentials()
        self.client = get_client()
        self.project_id = os.environ['PROJECT_ID']
        self.client.set.default_project(self.project_id)

    def test_01_publish_model(self):
        TestWMLClientWithScikitLearn.logger.info("Creating scikit-learn model ...")

        model_data = create_scikit_learn_model_data()
        predicted = model_data['prediction']

        TestWMLClientWithScikitLearn.logger.debug(predicted)
        self.assertIsNotNone(predicted)

        self.logger.info("Publishing scikit-learn model ...")

        self.client.repository.ModelMetaNames.show()

        model_props = {self.client.repository.ModelMetaNames.NAME: "ScikitModel",
            self.client.repository.ModelMetaNames.TYPE: "scikit-learn_0.20",
            self.client.repository.ModelMetaNames.RUNTIME_UID: "scikit-learn_0.20-py3.6"
                       }
        published_model_details = self.client.repository.store_model(model=model_data['model'], meta_props=model_props, training_data=model_data['training_data'], training_target=model_data['training_target'])
        TestWMLClientWithScikitLearn.model_uid = self.client.repository.get_model_uid(published_model_details)
       # TestWMLClientWithScikitLearn.model_url = self.client.repository.get_model_href(published_model_details)
        self.logger.info("Published model ID:" + str(TestWMLClientWithScikitLearn.model_uid))
        #self.logger.info("Published model URL:" + str(TestWMLClientWithScikitLearn.model_url))
        self.assertIsNotNone(TestWMLClientWithScikitLearn.model_uid)
        #self.assertIsNotNone(TestWMLClientWithScikitLearn.model_url)

    def test_03_download_model(self):
        TestWMLClientWithScikitLearn.logger.info("Download model")
        try:
            os.remove('download_test_url')
        except OSError:
            pass

        try:
            file = open('download_test_uid', 'r')
        except IOError:
            file = open('download_test_uid', 'w')
            file.close()

        self.client.repository.download(TestWMLClientWithScikitLearn.model_uid, filename='download_test_url')
        try:
            os.remove('test_ai_function.tar.gz')
        except:
            pass
    def test_04_get_details(self):
        TestWMLClientWithScikitLearn.logger.info("Get model details")
        details = self.client.repository.get_details(self.model_uid)
        TestWMLClientWithScikitLearn.logger.debug("Model details: " + str(details))
        self.assertTrue("ScikitModel" in str(details))

    def test_07_delete_model(self):
        TestWMLClientWithScikitLearn.logger.info("Delete model")
        self.client.repository.delete(TestWMLClientWithScikitLearn.model_uid)
       # self.client.spaces.delete(TestWMLClientWithScikitLearn.space_id)


if __name__ == '__main__':
    unittest.main()


#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import abc
from os import environ

import unittest
import time

from sklearn.pipeline import Pipeline

from ibm_watson_machine_learning import APIClient
from ibm_watson_machine_learning.experiment import AutoAI
from ibm_watson_machine_learning.deployment import WebService, Batch
from ibm_watson_machine_learning.workspace import WorkSpace
from ibm_watson_machine_learning.helpers.connections import DataConnection
from ibm_watson_machine_learning.experiment.autoai.optimizers import RemoteAutoPipelines
from ibm_watson_machine_learning.tests.utils import (get_wml_credentials, get_cos_credentials, get_space_id,
                                                     is_cp4d)
from ibm_watson_machine_learning.tests.utils.cleanup import space_cleanup, delete_model_deployment
from ibm_watson_machine_learning.utils.autoai.enums import PredictionType, RunStateTypes

from ibm_watson_machine_learning.tests.utils.assertions import get_and_predict_all_pipelines_as_lale, \
    validate_autoai_experiment

from ibm_watson_machine_learning.utils.autoai.utils import chose_model_output


class AbstractTestAutoAISyncNotebookFlow(abc.ABC):
    """
    The abstract tests which covers:
    - training AutoAI model on a dataset
    - autogenerated notebook scenario

    In order to execute test connection definitions must be provided
    in inheriting classes.
    """

    bucket_name = environ.get('BUCKET_NAME', "wml-autoaitests-qa")
    pod_version = environ.get('KB_VERSION', None)
    space_name = environ.get('SPACE_NAME', 'regression_tests_sdk_space')

    cos_endpoint = "https://s3.us.cloud-object-storage.appdomain.cloud"
    results_cos_path = 'results_wml_autoai'

    required_experiment_metadata_fields = ['prediction_type', 'prediction_column', 'holdout_size', 'scoring',
                                           'csv_separator', 'random_state', 'max_number_of_estimators',
                                           'training_data_reference', 'training_result_reference']

    experiment_metadata_fields_to_be_updated_after_training = ['training_data_reference', 'training_result_reference',
                                                               'holdout_size', 'scoring', 'csv_separator',
                                                               'random_state', 'max_number_of_estimators']

    # to be set in every child class:
    OPTIMIZER_NAME = "AutoAI regression test"

    SPACE_ONLY = True
    HISTORICAL_RUNS_CHECK = True

    experiment_info = dict(name=OPTIMIZER_NAME,
                           desc='test description',
                           prediction_type=PredictionType.MULTICLASS,
                           prediction_column='species',
                           autoai_pod_version=pod_version
                           )

    experiment_metadata: dict = None

    wml_client: 'APIClient' = None
    experiment: 'AutoAI' = None
    remote_auto_pipelines: 'RemoteAutoPipelines' = None
    wml_credentials = None
    cos_credentials = None
    pipeline_opt: 'RemoteAutoPipelines' = None
    service: 'WebService' = None
    service_batch: 'Batch' = None

    cos_resource_instance_id = None
    experiment_info: dict = None

    best_pipeline_name = False
    trained_pipeline_details = None
    run_id = None
    prev_run_id = None
    data_connection = None
    results_connection = None
    train_data = None

    pipeline: 'Pipeline' = None
    lale_pipeline = None
    deployed_pipeline = None
    hyperopt_pipelines = None
    new_pipeline = None
    new_sklearn_pipeline = None
    X_train = None
    y_train = None
    X_holdout = None
    y_holdout = None

    project_id = None
    space_id = None

    asset_id = None
    connection_id = None

    @classmethod
    def setUpClass(cls) -> None:
        """
        Load WML credentials from config.ini file based on ENV variable.
        """
        cls.wml_credentials = get_wml_credentials()
        cls.wml_client = APIClient(wml_credentials=cls.wml_credentials.copy())

        cls.cos_credentials = get_cos_credentials()
        cls.cos_endpoint = cls.cos_credentials.get('endpoint_url')
        cls.cos_resource_instance_id = cls.cos_credentials.get('resource_instance_id')

        cls.project_id = cls.wml_credentials.get('project_id')

    def test_00a_space_cleanup(self):
        space_checked = False
        while not space_checked:
            space_cleanup(self.wml_client,
                          get_space_id(self.wml_client, self.space_name,
                                       cos_resource_instance_id=self.cos_resource_instance_id),
                          days_old=7)
            space_id = get_space_id(self.wml_client, self.space_name,
                                    cos_resource_instance_id=self.cos_resource_instance_id)
            try:
                self.assertIsNotNone(space_id, msg="space_id is None")
                space_checked = True
            except AssertionError:
                space_checked = False

        AbstractTestAutoAISync.space_id = space_id

        if self.SPACE_ONLY:
            self.wml_client.set.default_space(self.space_id)
        else:
            self.wml_client.set.default_project(self.project_id)

    def test_01_initialize_AutoAI_experiment__pass_credentials__object_initialized(self):
        if self.SPACE_ONLY:
            AbstractTestAutoAISync.experiment_run_notebook = AutoAI(wml_credentials=self.wml_credentials.copy(),
                                                                    space_id=self.space_id)
        else:
            AbstractTestAutoAISync.experiment_run_notebook = AutoAI(wml_credentials=self.wml_credentials.copy(),
                                                                    project_id=self.project_id)

        self.assertIsInstance(self.experiment_run_notebook, AutoAI, msg="Experiment is not of type AutoAI.")

    @abc.abstractmethod
    def test_02_data_reference_setup(self):
        pass

    @abc.abstractmethod
    def test_02b_experiment_metadata_setup(self):
        pass

    def test_03_initialize_optimizer(self):
        AbstractTestAutoAISync.experiment_info = validate_autoai_experiment(self.experiment_info, self.pod_version)

        AbstractTestAutoAISync.remote_auto_pipelines = self.experiment_run_notebook.optimizer(
            **AbstractTestAutoAISync.experiment_info)

        self.assertIsInstance(self.remote_auto_pipelines, RemoteAutoPipelines,
                              msg="experiment.optimizer did not return RemoteAutoPipelines object")

    def test_04_get_configuration_parameters_of_remote_auto_pipeline(self):
        parameters = self.remote_auto_pipelines.get_params()
        print(parameters)

        self.assertIsInstance(parameters, dict, msg='Config parameters are not a dictionary instance.')

    def test_05_fit_run_training_of_auto_ai_in_wml(self):
        AbstractTestAutoAISync.trained_pipeline_details = self.remote_auto_pipelines.fit(
            training_data_reference=[self.data_connection],
            training_results_reference=self.results_connection,
            background_mode=True)

        AbstractTestAutoAISync.run_id = self.trained_pipeline_details['metadata']['id']
        self.assertIsNotNone(self.data_connection.auto_pipeline_params,
                             msg='DataConnection auto_pipeline_params was not updated.')

    def test_06a_get_run_status(self):
        status = self.remote_auto_pipelines.get_run_status()
        self.assertNotEqual(status, RunStateTypes.COMPLETED)

    def test_06b_get_run_details(self):
        parameters = self.remote_auto_pipelines.get_run_details()
        # print(parameters)
        self.assertIsNotNone(parameters)
        self.assertIn(self.run_id, str(parameters))

    def test_07_wait_for_first_pipeline(self):

        print(f"Run status = {self.remote_auto_pipelines.get_run_status()}")
        # note: check if first pipeline was generated

        metrics = self.wml_client.training.get_details(self.run_id)['entity']['status'].get('metrics', [])
        while chose_model_output("1") not in str(metrics) and self.remote_auto_pipelines.get_run_status() not in [
            'failed', 'canceled']:
            time.sleep(5)
            print(".", end=" ")
            metrics = self.wml_client.training.get_details(self.run_id)['entity']['status'].get('metrics', [])
        # end note

        status = self.remote_auto_pipelines.get_run_status()
        run_details = self.remote_auto_pipelines.get_run_details().get('entity')
        self.assertNotIn(status, ['failed', 'canceled'], msg=f"Training finished with status {status}. \n"
                                                             f"Details: {run_details.get('status')}")
        print("\n 1st pipeline completed")

    def test_08_update_experiment_metadata_data_referances(self):
        pipeline_parameters = self.remote_auto_pipelines.get_params()
        for param_to_add in self.experiment_metadata_fields_to_be_updated_after_training:
            if param_to_add == 'training_data_reference':
                self.experiment_metadata[param_to_add] = self.remote_auto_pipelines.get_data_connections()
            elif param_to_add == 'test_data_references':
                self.experiment_metadata[param_to_add] = self.remote_auto_pipelines.get_data_connections()
            elif param_to_add == 'training_result_reference':
                results_reference = self.remote_auto_pipelines.get_run_details()['entity']['results_reference']
                self.experiment_metadata[param_to_add] = DataConnection._from_dict(results_reference)
            if param_to_add not in self.experiment_metadata:
                self.experiment_metadata[param_to_add] = pipeline_parameters.get(param_to_add)

    def test_09_validate_experiment_metadata(self):
        missing_params = []
        for r_parameter in self.required_experiment_metadata_fields:
            if r_parameter not in self.experiment_metadata:
                missing_params.append(r_parameter)

        self.assertEqual(len(missing_params), 0, f"Missing in experiment metadata: {missing_params}")

    ###########################################
    #             NOTEBOOK CELLS              #
    ###########################################

    def test_10_notebook_wml_credential_and_experiment(self):
        wml_credentials = {
            "apikey": self.wml_credentials.get('apikey'),
            "url": self.experiment_metadata['deployment_url']
        }

        self.pipeline_optimizer = AutoAI(wml_credentials,
                                         project_id=self.experiment_metadata['project_id']).runs.get_optimizer(
            metadata=self.experiment_metadata)

    def test_10b_download_data_with_holdout_not_in_experiment_notebook(self):
        AbstractTestAutoAISync.X_train, AbstractTestAutoAISync.X_holdout, AbstractTestAutoAISync.y_train, AbstractTestAutoAISync.y_holdout = \
            self.pipeline_optimizer.get_data_connections()[0].read(
                with_holdout_split=True)

        print("train data sample:")
        print(AbstractTestAutoAISync.X_train)
        print(AbstractTestAutoAISync.y_train)
        print("holdout data sample:")
        print(AbstractTestAutoAISync.X_holdout)
        print(AbstractTestAutoAISync.y_holdout)

        self.assertGreater(len(AbstractTestAutoAISync.X_train), 0)
        self.assertGreater(len(AbstractTestAutoAISync.X_holdout), 0)
        self.assertEqual(len(AbstractTestAutoAISync.X_train), len(AbstractTestAutoAISync.y_train))
        self.assertEqual(len(AbstractTestAutoAISync.X_holdout), len(AbstractTestAutoAISync.y_holdout))

    def test_11_get_params(self):
        parameters = self.pipeline_optimizer.get_params()
        print(parameters)

        self.assertIsInstance(parameters, dict, msg='Config parameters are not a dictionary instance.')

    def test_12_get_summary(self):
        summary = self.pipeline_optimizer.summary()
        self.best_pipeline_name = list(summary.index)[0]

        print(summary)

        self.assertGreater(len(summary), 0,
                           msg=f"Summary DataFrame is empty. ")

        # check if pipelines are not duplicated
        self.assertEqual(len(summary.index.unique()), len(summary),
                         msg="Some pipeline names are duplicated in the summary")

    def test_13_pipeline_model(self):
        self.pipeline_model = self.pipeline_optimizer.get_pipeline()
        from lale.operators import TrainablePipeline
        self.assertIsInstance(self.lale_pipeline, TrainablePipeline,
                              msg="Fetched pipeline is not of TrainablePipeline instance.")

        feature_importance = self.pipeline_optimizer.get_pipeline_details()['features_importance']

        from pandas import DataFrame

        self.assertIsInstance(feature_importance, DataFrame,
                              msg="Feature importance is not returned as pandas.DataFrame.")

        self.assertGreater(len(feature_importance), 0,
                           msg=f"feature_importance DataFrame is empty.")

    def test_14a_pipeline_pretty_print(self):
        pipeline_code = self.pipeline_model.pretty_print(combinators=False)

        global pipeline
        exception = None
        try:
            exec(pipeline_code, globals())
        except Exception as exception:
            self.assertIsNone(exception, msg="Pretty print from lale pipeline was not successful")
        pipeline_str = str(pipeline.export_to_sklearn_pipeline()).replace("pipelinePipeline", "Pipeline")
        scikit_pipeline_str = str(self.pipeline_model.export_to_sklearn_pipeline())
        print(f"Pretty_print pipeline {pipeline_str}\n")
        print(f"Client pipeline {scikit_pipeline_str}")

        self.assertEqual(pipeline_str, scikit_pipeline_str, msg=f"Code is not equal, pretty_print_score:\
        {scikit_pipeline_str}, pipeline score {pipeline_str}")

    @unittest.skip
    def test_14b_pipeline_pretty_print_check_if_generated_python_pipeline_code_is_correct(self):
        variables = {"X_train": self.X_train.values,
                     "X_test": self.X_test.values,
                     "y_train": self.y_train.values,
                     "y_test": self.y_test.values
                     }
        global pretty_print_score
        global pipeline

        base_string: str = "\
        import pandas as pd\r\n\
        from sklearn.metrics import r2_score"
        end_string = "\
        pipeline.fit(X_train,y_train)\r\n\
        pretty_print_score = r2_score(y_test,pipeline.predict(X_test))"

        pipeline_code = self.pipeline_model.pretty_print(combinators=False)

        from sklearn.metrics import roc_auc_score

        scikit_pipeline = self.pipeline_model.export_to_sklearn_pipeline()
        scikit_pipeline.fit(self.X_train.values, self.y_train.values)
        print(f"scikit_pipeline_params{scikit_pipeline.get_params(deep=True)}")
        pipeline_score = roc_auc_score(self.y_test.values, scikit_pipeline.predict(self.X_test.values))

        exception = None
        try:
            # print(base_string + "\r\n" + pipeline_code + "\r\n" + end_string)
            exec(base_string + "\r\n" + pipeline_code + "\r\n" + end_string, variables, globals())
            pipeline = pipeline.export_to_sklearn_pipeline()
            print(f"pretty_pipeline_params{pipeline.get_params(deep=True)}")

        except Exception as exception:
            self.assertIsNone(exception, msg="Pretty print from lale pipeline was not successful")
        print(pipeline_score)
        print(pretty_print_score)
        self.assertAlmostEqual(first=pretty_print_score, second=pipeline_score, places=1, msg=f"Scores are different at first decimal place:\
                {pretty_print_score}, pipeline score {pipeline_score}")

    ###############################
    #         DEPLOYMENT          #
    ###############################

    def test_20_deployment_creation(self):
        target_space_id = self.space_id

        from ibm_watson_machine_learning.deployment import WebService

        if self.SPACE_ONLY:
            self.assertIsNotNone(self.experiment_metadata.get('space_id'), msg="Missing space id in metadata")
            AbstractTestAutoAISync.service = WebService(
                source_wml_credentials=self.wml_credentials,
                target_wml_credentials=self.wml_credentials,
                source_project_id=self.experiment_metadata['space_id'],
                target_space_id=target_space_id
            )
        else:
            AbstractTestAutoAISync.self.assertIsNotNone(self.experiment_metadata.get('project_id'),
                                                        msg="Missing project id in metadata")
            service = WebService(
                source_wml_credentials=self.wml_credentials,
                target_wml_credentials=self.wml_credentials,
                source_project_id=self.experiment_metadata['project_id'],
                target_space_id=target_space_id
            )

        AbstractTestAutoAISync.service.create(
            model=self.best_pipeline_name,
            metadata=self.experiment_metadata,
            deployment_name='Best_pipeline_webservice'
        )
        print(AbstractTestAutoAISync.service)

    def test_21_webservice_get_params(self):
        params = AbstractTestAutoAISync.service.get_params()
        print(params)
        self.assertIsNotNone(params)

    def test_22_webservice_score_markdown_in_experiment_notebook(self):
        """
        MARKDOWN:
        If you want to work with the web service in an external Python application, follow these steps to retrieve the service object:

        Initialize the service by service = WebService(target_wml_credentials=wml_credentials,target_space_id=experiment_metadata['space_id'])
        Get deployment_id by service.list() method
        Get webservice object by service.get('deployment_id') method

        After that you can call service.score(score_records_df) method. The score() method accepts pandas.DataFrame object.
        """
        score_records_df = self.X_holdout

        predictions = self.service.score(score_records_df)

        print(predictions)
        self.assertIsNotNone(predictions)
        self.assertEqual(len(predictions['predictions'][0]['values']), len(score_records_df))

    ########
    # LALE #
    ########

    def test_31__get_pipeline__load_lale_pipeline__pipeline_loaded(self):
        AbstractTestAutoAISync.lale_pipeline = self.remote_auto_pipelines.get_pipeline()
        print(f"Fetched pipeline type: {type(self.lale_pipeline)}")

        from lale.operators import TrainablePipeline
        self.assertIsInstance(self.lale_pipeline, TrainablePipeline,
                              msg="Fetched pipeline is not of TrainablePipeline instance.")
        predictions = self.lale_pipeline.predict(
            X=self.X_values[:5])
        print(predictions)

    def test_32_get_all_pipelines_as_lale(self):
        get_and_predict_all_pipelines_as_lale(self.remote_auto_pipelines, self.X_values)



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest

import ibm_boto3
from ibm_watson_machine_learning.experiment import AutoAI
from ibm_watson_machine_learning.helpers.connections import DataConnection, S3Location
from ibm_watson_machine_learning.utils.autoai.errors import WMLClientError
from ibm_watson_machine_learning.tests.utils import bucket_exists, create_bucket, is_cp4d
from ibm_watson_machine_learning.tests.autoai.abstract_tests_classes import AbstractTestTSAsync, \
    AbstractTestWebservice

from ibm_watson_machine_learning.utils.autoai.enums import PredictionType, Metrics


class TestAutoAIRemote(AbstractTestTSAsync, unittest.TestCase):
    """
    The test can be run on Cloud only
    """

    cos_resource = None
    data_location = './autoai/data/Twitter_volume_AMZN.csv'

    data_cos_path = 'Twitter_volume_AMZN.csv'

    batch_payload_location = data_location
    batch_payload_cos_location = data_cos_path

    SPACE_ONLY = True
    BATCH_DEPLOYMENT = False

    OPTIMIZER_NAME = "Twitter_volume_AMZN test sdk"
    DEPLOYMENT_NAME = OPTIMIZER_NAME + "Deployment"

    target_space_id = None

    experiment_info = dict(
        name=OPTIMIZER_NAME,
        prediction_type=PredictionType.FORECASTING,
        prediction_columns=['value'],
        timestamp_column_name='timestamp',
        backtest_num=4,
        lookback_window=5,
        forecast_window=2,
        holdout_size=0.05,
        max_number_of_estimators=1,
        notebooks=True
    )

    def test_00d_prepare_data_asset(self):
        asset_details = self.wml_client.data_assets.create(
            name=self.data_location.split('/')[-1],
            file_path=self.data_location)

        TestAutoAIRemote.asset_id = self.wml_client.data_assets.get_id(asset_details)
        self.assertIsInstance(self.asset_id, str)

    def test_02_data_reference_setup(self):
        TestAutoAIRemote.data_connection = DataConnection(data_asset_id=self.asset_id)
        TestAutoAIRemote.results_connection = None

        self.assertIsNotNone(obj=TestAutoAIRemote.data_connection)
        self.assertIsNone(obj=TestAutoAIRemote.results_connection)

    def test_09a_predict_using_fitted_pipeline(self):
        predictions = self.remote_auto_pipelines.predict(X=self.scoring_df[:5].values)
        print(predictions)
        self.assertGreater(len(predictions), 0)

    def test_09ab_predict_using_fitted_pipeline(self):
        predictions = self.remote_auto_pipelines.predict(X=self.scoring_df[:5].values)
        print(predictions)
        self.assertGreater(len(predictions), 0)

    def test_09b_get_pipeline_optimizer_from_experiment_metdata(self):
        from ibm_watson_machine_learning.helpers import ContainerLocation

        training_data_references = [
            DataConnection(
                data_asset_id=self.asset_id
            ),
        ]
        training_result_reference = DataConnection(
            location=ContainerLocation(
                path=f'default_autoai_out/{self.run_id}/data/automl',
                model_location=f'default_autoai_out/{self.run_id}/data/automl/model.zip',
                training_status=f'default_autoai_out/{self.run_id}/training-status.json'
            )
        )

        experiment_metadata = dict(
            prediction_type='forecasting',
            prediction_columns=['value'],
            csv_separator=',',
            holdout_size=0.05,
            training_data_reference=training_data_references,
            training_result_reference=training_result_reference,
            timestamp_column_name='timestamp',
            backtest_num=4,
            lookback_window=5,
            forecast_window=2,
            max_num_daub_ensembles=1,
            deployment_url='https://yp-qa.ml.cloud.ibm.com',
            space_id=self.space_id
        )
        summary = self.remote_auto_pipelines.summary()
        best_pipeline_name = list(summary.index)[0]

        from ibm_watson_machine_learning.deployment import WebService

        service = WebService(
            source_wml_credentials=self.wml_credentials,
            target_wml_credentials=self.wml_credentials,
            source_space_id=experiment_metadata['space_id'],
            target_space_id=experiment_metadata['space_id']
        )
        service.create(
            model=best_pipeline_name,
            metadata=experiment_metadata,
            deployment_name='Best_pipeline_webservice'
        )

        self.assertIsNotNone(service)

    def test_09c_get_pipeline_optimizer_from_experiment_metdata(self):
        from ibm_watson_machine_learning.helpers import ContainerLocation

        training_data_references = [
            DataConnection(
                data_asset_id=self.asset_id
            ),
        ]
        training_result_reference = DataConnection(
            location=ContainerLocation(
                path=f'default_autoai_out/{self.run_id}/data/automl',
                model_location=f'default_autoai_out/{self.run_id}/data/automl/model.zip',
                training_status=f'default_autoai_out/{self.run_id}/training-status.json'
            )
        )

        experiment_metadata = dict(
            prediction_type='forecasting',
            prediction_columns=['value'],
            csv_separator=',',
            holdout_size=0.05,
            training_data_reference=training_data_references,
            training_result_reference=training_result_reference,
            timestamp_column_name='timestamp',
            backtest_num=4,
            lookback_window=5,
            forecast_window=2,
            max_num_daub_ensembles=1,
            deployment_url='https://yp-qa.ml.cloud.ibm.com',
            space_id=self.space_id
        )
        pipeline_optimizer = AutoAI().runs.get_optimizer(metadata=experiment_metadata,
                                                         wml_client=self.remote_auto_pipelines._workspace.wml_client)

        self.assertIsNotNone(pipeline_optimizer)

    def test_99_delete_data_asset(self):
        if not self.SPACE_ONLY:
            self.wml_client.set.default_project(self.project_id)

        self.wml_client.data_assets.delete(self.asset_id)

        with self.assertRaises(WMLClientError):
            self.wml_client.data_assets.get_details(self.asset_id)


if __name__ == '__main__':
    unittest.main()



"""
    IBM Watson Machine Learning REST API

    ## Authorization  ### IBM Watson Machine Learning Credentials  To start working with API one needs to generate an `access token` using the `username` and `password` available on the Service Credentials tab of the IBM Watson Machine Learning service instance or also available in the VCAP environment variable.  Example of the Service Credentials:  ```json {     \"url\": \"https://ibm-watson-ml.mybluemix.net\",     \"access_key\": \"ERY9vcBfE4sE+F4g8hcotF9L+j81WXWeZv\",     \"username\": \"c1ef4b80-2ee2-458e-ab92-e9ca97ec657d\",     \"password\": \"030528d4-5a3e-4d4c-9258-5d553513be6f\" } ```  Example of obtaining `access token` from Token Endpoint using HTTP Basic Auth (for details please refer to Token section below):  ` curl --basic --user username:password https://ibm-watson-ml.mybluemix.net/v3/identity/token `   ### Apache Spark Service Credentials  The IBM Watson Machine Learning co-operates with the Apache Spark as a Service to deploy pipeline models. For API methods requiring Apache Spark Service instance a custom header: `X-Spark-Service-Instance` with Service Credentials must be specified. The header value is a **base64 encoded** string with the Service Credentials JSON data.  [Example of API method requiring Apache Spark Service](https://console.ng.bluemix.net/docs/services/PredictiveModeling/index-gentopic1.html#pm_service_api_spark_batch)

    OpenAPI spec version: 2.0.0
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import absolute_import

# import models into sdk package
from .models.array_model_metrics_output import ArrayModelMetricsOutput
from .models.array_model_output import ArrayModelOutput
from .models.array_model_version_output import ArrayModelVersionOutput
from .models.array_pipeline_output import ArrayPipelineOutput
from .models.array_pipeline_version_output import ArrayPipelineVersionOutput
from .models.artifact_author import ArtifactAuthor
from .models.artifact_version_metadata import ArtifactVersionMetadata
from .models.artifact_version_short_metadata import ArtifactVersionShortMetadata
from .models.batch_deploy_output import BatchDeployOutput
from .models.batch_deploy_output_entity import BatchDeployOutputEntity
from .models.batch_deploy_output_entity_execution import BatchDeployOutputEntityExecution
from .models.batch_deploy_output_meta import BatchDeployOutputMeta
from .models.batch_input import BatchInput
from .models.batch_output import BatchOutput
from .models.batch_output_array import BatchOutputArray
from .models.cols_output import ColsOutput
from .models.connection import Connection
from .models.deploy_input import DeployInput
from .models.error_message import ErrorMessage
from .models.evaluation_definition import EvaluationDefinition
from .models.evaluation_definition_metrics import EvaluationDefinitionMetrics
from .models.input_data_schema import InputDataSchema
from .models.internal_input_batch import InternalInputBatch
from .models.internal_output_batch import InternalOutputBatch
from .models.json_patch_array import JsonPatchArray
from .models.json_patch_entity import JsonPatchEntity
from .models.meta_object import MetaObject
from .models.meta_object_metadata import MetaObjectMetadata
from .models.model_input import ModelInput
from .models.model_metrics import ModelMetrics
from .models.model_metrics_values import ModelMetricsValues
from .models.model_output import ModelOutput
from .models.model_output_entity import ModelOutputEntity
from .models.model_output_entity_pipeline_version import ModelOutputEntityPipelineVersion
from .models.model_training_data_ref import ModelTrainingDataRef
from .models.model_type import ModelType
from .models.model_version_input import ModelVersionInput
from .models.model_version_output import ModelVersionOutput
from .models.model_version_output_entity import ModelVersionOutputEntity
from .models.model_version_output_entity_model import ModelVersionOutputEntityModel
from .models.online_deploy_output import OnlineDeployOutput
from .models.online_output import OnlineOutput
from .models.online_output_array import OnlineOutputArray
from .models.pipeline_input import PipelineInput
from .models.pipeline_output import PipelineOutput
from .models.pipeline_output_entity import PipelineOutputEntity
from .models.pipeline_type import PipelineType
from .models.pipeline_version_input import PipelineVersionInput
from .models.pipeline_version_output import PipelineVersionOutput
from .models.pipeline_version_output_entity import PipelineVersionOutputEntity
from .models.pipeline_version_output_entity_parent import PipelineVersionOutputEntityParent
from .models.runtime_environment import RuntimeEnvironment
from .models.score_input import ScoreInput
from .models.score_output import ScoreOutput
from .models.spark_service import SparkService
from .models.stream_input_internal import StreamInputInternal
from .models.stream_internal import StreamInternal
from .models.stream_output import StreamOutput
from .models.stream_output_array import StreamOutputArray
from .models.stream_output_internal import StreamOutputInternal
from .models.token_response import TokenResponse
from .models.training_data_schema import TrainingDataSchema
from .models.output_data_schema import OutputDataSchema
from .models.model_content_location import ModelContentLocation
from .models.training_models import TrainingModels
from .models.space_models import SpaceModels
from .models.runtime_models import RuntimeModels
from .models.software_spec_models import SoftwareSpecModels
from .models.model_schemas import ModelSchemas
from .models.pipeline_models import PipelinesModels
from .models.model_content_location import ModelContentLocation
from .models.custom_models import ModelsCustom
from .models.metrics_models import ModelsMetrics
from .models.model_schemas import ModelSchemas
from .models.size_models import ModelsSize
from .models.training_models import TrainingModels
from .models.model_definition_models import ModelDefinitionModels




# import v3 models into sdk package
from .models.array_data_input_repository import ArrayDataInputRepository
from .models.artifact_version_repository import ArtifactVersionRepository
from .models.author_repository import AuthorRepository
from .models.connection_object_with_name_repository import ConnectionObjectWithNameRepository
from .models.error_repository import ErrorRepository
from .models.error_repository_target import ErrorRepositoryTarget
from .models.error_schema_repository import ErrorSchemaRepository
from .models.evaluation_definition_repository import EvaluationDefinitionRepository
from .models.evaluation_definition_repository_metrics import EvaluationDefinitionRepositoryMetrics
from .models.framework_output_repository import FrameworkOutputRepository
from .models.framework_output_repository_libraries import FrameworkOutputRepositoryLibraries
from .models.framework_output_repository_runtimes import  FrameworkOutputRepositoryRuntimes
from .models.meta_object_repository import MetaObjectRepository
from .models.meta_object_repository_metadata import MetaObjectRepositoryMetadata
from .models.ml_assets_create_experiment_input import MlAssetsCreateExperimentInput
from .models.ml_assets_create_experiment_output import MlAssetsCreateExperimentOutput
from .models.ml_assets_create_experiment_output_array import MlAssetsCreateExperimentOutputArray
from .models.ml_assets_create_model_input import MlAssetsCreateModelInput
from .models.ml_assets_create_model_output import MlAssetsCreateModelOutput
from .models.ml_assets_create_model_output_array import MlAssetsCreateModelOutputArray
from .models.content_location import ContentLocation
from .models.content_status import ContentStatus
from .models.ml_assets_model_size_output import MlAssetsModelSizeOutput
from .models.ml_assets_get_presigned_url_output import MlAssetsGetPreSignedUrlOutput
from .models.ml_assets_upload_content_output import MlAssetsUploadContentOutput
from .models.ml_assets_upload_content_output_metadata import MlAssetsUploadContentOutputMetadata
from .models.ml_assets_upload_content_output_entity import MlAssetsUploadContentOutputEntity



# import models into sdk package
from .models.array_model_version_metrics_experiments import ArrayModelVersionMetricsExperiments
from .models.array_training_output_experiments import ArrayTrainingOutputExperiments
from .models.author_experiments import AuthorExperiments
from .models.compute_configuration_experiments import ComputeConfigurationExperiments
from .models.connection_object_source_experiments import ConnectionObjectSourceExperiments
from .models.connection_object_target_experiments import ConnectionObjectTargetExperiments
from .models.error_experiments import ErrorExperiments
from .models.error_experiments_target import ErrorExperimentsTarget
from .models.error_schema_experiments import ErrorSchemaExperiments
from .models.evaluation_definition_experiments import EvaluationDefinitionExperiments
from .models.tag_repository import TagRepository
from .models.experiment_input import ExperimentInput
from .models.experiment_input_settings import ExperimentInputSettings
from .models.experiment_output_array import ExperimentOutputArray
from .models.experiment_output_array_first import ExperimentOutputArrayFirst
from .models.experiment_output import ExperimentOutput
from .models.experiment_patch import ExperimentPatch
from .models.experiment_status_experiments import ExperimentStatusExperiments
from .models.hyper_parameters_experiments import HyperParametersExperiments
from .models.hyper_parameters_experiments_inner_values_range import HyperParametersExperimentsInnerValuesRange
from .models.hyper_parameters_for_status_experiments import HyperParametersForStatusExperiments
from .models.hyper_parameters_for_status_experiments_inner import HyperParametersForStatusExperimentsInner
from .models.hyper_parameters_optimization_experiments import HyperParametersOptimizationExperiments
from .models.meta_object_experiments import MetaObjectExperiments
from .models.meta_object_experiments_metadata import MetaObjectExperimentsMetadata
from .models.metric_object_experiments import MetricObjectExperiments
from .models.model_version_metrics_experiments import ModelVersionMetricsExperiments
from .models.patch_operation_experiments import PatchOperationExperiments
from .models.training_output_experiments import TrainingOutputExperiments
from .models.training_reference_experiments import TrainingReferenceExperiments
from .models.training_status_experiments import TrainingStatusExperiments
from .models.training_status_experiments_result import TrainingStatusExperimentsResult
from .models.hyper_parameters import HyperParameters
from .models.hyper_parameters_optimization_experiments_method import HyperParametersOptimizationExperimentsMethod
from .models.hyper_parameters_optimization_experiments_method_parameters import HyperParametersOptimizationExperimentsMethodParameters
from .models.hyper_parameters_experiments_double_range import HyperParametersExperimentsDoubleRange
from .models.hyper_parameters_experiments_int_range import HyperParametersExperimentsIntRange














# import apis into sdk package
from .apis.repository_api import RepositoryApi
from .apis.token_api import TokenApi

# import ApiClient
from .api_client import ApiClient

from .configuration import Configuration

configuration = Configuration()



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import print_function
import requests
from ibm_watson_machine_learning.utils import SPACES_DETAILS_TYPE, INSTANCE_DETAILS_TYPE, MEMBER_DETAILS_TYPE,SW_SPEC_DETAILS_TYPE, STR_TYPE, STR_TYPE_NAME, docstring_parameter, meta_props_str_conv, str_type_conv, get_file_from_cos
from ibm_watson_machine_learning.metanames import SwSpecMetaNames
from ibm_watson_machine_learning.wml_resource import WMLResource
from ibm_watson_machine_learning.wml_client_error import WMLClientError, ApiRequestFailure
import os,json

_DEFAULT_LIST_LENGTH = 50


class SwSpec(WMLResource):
    """
    Store and manage your software specs.

    """
    ConfigurationMetaNames = SwSpecMetaNames()
    """MetaNames for Software Specification creation."""
    def __init__(self, client):
        WMLResource.__init__(self, __name__, client)
        self._ICP = client.ICP
        self.software_spec_list = None
        if self._client.WSD_20:
            self.software_spec_list = {
                                "default_py3.6": "0062b8c9-8b7d-44a0-a9b9-46c416adcbd9",
                                "scikit-learn_0.20-py3.6": "09c5a1d0-9c1e-4473-a344-eb7b665ff687",
                                "ai-function_0.1-py3.6": "0cdb0f1e-5376-4f4d-92dd-da3b69aa9bda",
                                "shiny-r3.6": "0e6e79df-875e-4f24-8ae9-62dcc2148306",
                                "pytorch_1.1-py3.6": "10ac12d6-6b30-4ccd-8392-3e922c096a92" ,
                                "scikit-learn_0.22-py3.6": "154010fa-5b3b-4ac1-82af-4d5ee5abbc85",
                                "default_r3.6":  "1b70aec3-ab34-4b87-8aa0-a4a3c8296a36",
                                "tensorflow_1.15-py3.6":  "2b73a275-7cbf-420b-a912-eae7f436e0bc",
                                "pytorch_1.2-py3.6":  "2c8ef57d-2687-4b7d-acce-01f94976dac1",
                                "spark-mllib_2.3":  "2e51f700-bca0-4b0d-88dc-5c6791338875",
                                "pytorch-onnx_1.1-py3.6-edt": "32983cea-3f32-4400-8965-dde874a8d67e",
                                "spark-mllib_2.4": "390d21f8-e58b-4fac-9c55-d7ceda621326",
                                "xgboost_0.82-py3.6":  "39e31acd-5f30-41dc-ae44-60233c80306e",
                                "pytorch-onnx_1.2-py3.6-edt": "40589d0e-7019-4e28-8daa-fb03b6f4fe12",
                                "spark-mllib_2.4-r_3.6": "49403dff-92e9-4c87-a3d7-a42d0021c095",
                                "xgboost_0.90-py3.6": "4ff8d6c2-1343-4c18-85e1-689c965304d3",
                                "pytorch-onnx_1.1-py3.6":  "50f95b2a-bc16-43bb-bc94-b0bed208c60b",
                                "spark-mllib_2.4-scala_2.11": "55a70f99-7320-4be5-9fb9-9edb5a443af5",
                                "spss-modeler_18.1":  "5c3cad7e-507f-4b2a-a9a3-ab53a21dee8b",
                                "spark-mllib_2.3-r_3.6":  "6586b9e3-ccd6-4f92-900f-0f8cb2bd6f0c",
                                "spss-modeler_18.2":  "687eddc9-028a-4117-b9dd-e57b36f1efa5",
                                "pytorch-onnx_1.2-py3.6":  "692a6a4d-2c4d-45ff-a1ed-b167ee55469a",
                                "do_12.9":  "75a3a4b0-6aa0-41b3-a618-48b1f56332a6",
                                "spark-mllib_2.3-scala_2.11": "7963efe5-bbec-417e-92cf-0574e21b4e8d",
                                "caffe_1.0-py3.6":  "7bb3dbe2-da6e-4145-918d-b6d84aa93b6b",
                                "cuda-py3.6":  "82c79ece-4d12-40e6-8787-a7b9e0f62770",
                                "hybrid_0.1":  "8c1a58c6-62b5-4dc4-987a-df751c2756b6",
                                "caffe-ibm_1.0-py3.6":  "8d863266-7927-4d1e-97d7-56a7f4c0a19b",
                                "spss-modeler_17.1":  "902d0051-84bd-4af6-ab6b-8f6aa6fdeabb",
                                "do_12.10":  "9100fd72-8159-4eb9-8a0b-a87e12eefa36",
                                "hybrid_0.2":  "9b3f9040-9cee-4ead-8d7a-780600f542f7"}

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_details(self,sw_spec_uid=None):
        """
            Get software specification details. If no sw_spec_id is passed, details for all software specifications will be returned.

            **Parameters**

            .. important::
                **returns**: sw_spec UID\n
                **return type**: str

            **Output**

            .. important::
                #. **sw_spec_details**:  Metadata of the stored sw_spec\n
                   **type**: dict\n

            **Example**

             >>> sw_spec_details = client.software_specifications.get_details(sw_spec_uid)

        """
        if self._client.WSD_20:
            raise WMLClientError(u'get_details API is not supported in Watson Studio Desktop.')

        SwSpec._validate_type(sw_spec_uid, u'sw_spec_uid', STR_TYPE, False)

        if sw_spec_uid:
            if not self._ICP:
                response = requests.get(self._client.service_instance._href_definitions.get_sw_spec_href(sw_spec_uid),
                                        params=self._client._params(skip_space_project_chk=True),
                                        headers=self._client._get_headers())
            else:
                if self._client.ICP_PLATFORM_SPACES:
                    response = requests.get(
                        self._client.service_instance._href_definitions.get_sw_spec_href(sw_spec_uid),
                        params=self._client._params(skip_space_project_chk=True),
                        headers=self._client._get_headers(), verify=False)
                else:
                    response = requests.get(
                        self._client.service_instance._href_definitions.get_sw_spec_href(sw_spec_uid),
                        params=self._client._params(),
                        headers=self._client._get_headers(), verify=False)

            if response.status_code == 200:
                return self._get_required_element_from_response(
                    self._handle_response(200, u'get sw spec details', response))
            else:
                return self._handle_response(200, u'get sw spec details', response)
        else:
            if not self._ICP:
                response = requests.get(self._client.service_instance._href_definitions.get_sw_specs_href(),
                                        params=self._client._params(skip_space_project_chk=True),
                                        headers=self._client._get_headers())
            else:
                if self._client.ICP_PLATFORM_SPACES:
                    response = requests.get(
                        self._client.service_instance._href_definitions.get_sw_specs_href(),
                        params=self._client._params(skip_space_project_chk=True),
                        headers=self._client._get_headers(), verify=False)
                else:
                    response = requests.get(
                        self._client.service_instance._href_definitions.get_sw_specs_href(),
                        params=self._client._params(),
                        headers=self._client._get_headers(), verify=False)

            if response.status_code == 200:
                return {'resources': [self._get_required_element_from_response(x)
                                      for x in self._handle_response(200, u'get sw specs details', response)['resources']]}
            else:
                return self._handle_response(200, u'get sw specs details', response)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def store(self, meta_props):
        """
                Create a software specification.

                **Parameters**

                .. important::
                   #. **meta_props**:  meta data of the space configuration. To see available meta names use:\n
                                    >>> client.software_specifications.ConfigurationMetaNames.get()

                      **type**: dict\n

                **Output**

                .. important::

                    **returns**: metadata of the stored space\n
                    **return type**: dict\n

                **Example**

                 >>> meta_props = {
                 >>>    client.software_specifications.ConfigurationMetaNames.NAME: "skl_pipeline_heart_problem_prediction",
                 >>>    client.software_specifications.ConfigurationMetaNames.DESCRIPTION: "description scikit-learn_0.20",
                 >>>    client.software_specifications.ConfigurationMetaNames.PACKAGE_EXTENSIONS_UID: [],
                 >>>    client.software_specifications.ConfigurationMetaNames.SOFTWARE_CONFIGURATIONS: {},
                 >>>    client.software_specifications.ConfigurationMetaNames.BASE_SOFTWARE_SPECIFICATION_ID: "guid"
                 >>> }

        """
        WMLResource._chk_and_block_create_update_for_python36(self)
        if self._client.WSD_20:
            raise WMLClientError(u'store() API is not supported in Watson Studio Desktop.')

        SwSpec._validate_type(meta_props, u'meta_props', dict, True)
        sw_spec_meta = self.ConfigurationMetaNames._generate_resource_metadata(
            meta_props,
            with_validation=True,
            client=self._client)

        sw_spec_meta_json = json.dumps(sw_spec_meta)
        href = self._client.service_instance._href_definitions.get_sw_specs_href()

        if not self._ICP:
            creation_response = requests.post(href, params=self._client._params(), headers=self._client._get_headers(), data=sw_spec_meta_json)
        else:
            creation_response = requests.post(href, params=self._client._params(), headers=self._client._get_headers(), data=sw_spec_meta_json, verify=False)

        sw_spec_details = self._handle_response(201, u'creating sofware specifications', creation_response)

        return sw_spec_details


    def list(self, limit=None):
        """
           List software specifications.

           **Parameters**

           .. important::
                #. **limit**:  limit number of fetched records\n
                   **type**: int\n

           **Output**

           .. important::
                This method only prints the list of all software specifications in a table format.\n
                **return type**: None\n

           **Example**

            >>> client.software_specifications.list()
        """

        if not self._client.WSD_20:
            href = self._client.service_instance._href_definitions.get_sw_specs_href()
            if not self._ICP:
                response = requests.get(href, params=self._client._params(), headers=self._client._get_headers())
            else:
                response = requests.get(href, params=self._client._params(), headers=self._client._get_headers(), verify=False)
            self._handle_response(200, u'list sw_specs', response)
            asset_details = self._handle_response(200, u'list assets', response)["resources"]
            sw_spec_values = [
                (m[u'metadata'][u'name'], m[u'metadata'][u'asset_id'],m[u'entity'][u'software_specification'].get('type', 'derived')) for
                m in asset_details]

            self._list(sw_spec_values, [u'NAME', u'ASSET_ID', u'TYPE'], limit, _DEFAULT_LIST_LENGTH)
        else:
            from tabulate import tabulate
            header = [u'NAME', u'ASSET_ID', u'TYPE']
            print(tabulate(self.software_spec_list.items(), headers=header))


    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_id(sw_spec_details):
        """
                Get Unique Id of software specification.

                **Parameters**

                .. important::

                   #. **sw_spec_details**:  Metadata of the software specification\n
                      **type**: dict\n

                **Output**

                .. important::

                    **returns**: Unique Id of software specification \n
                    **return type**: str\n

                **Example**


                 >>> asset_uid = client.software_specifications.get_id(sw_spec_details)

        """

        return SwSpec.get_uid(sw_spec_details)

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_uid(sw_spec_details):
        """
                Get Unique Id of software specification. Deprecated!! Use get_id(sw_spec_details) instead

                **Parameters**

                .. important::

                   #. **sw_spec_details**:  Metadata of the software specification\n
                      **type**: dict\n

                **Output**

                .. important::

                    **returns**: Unique Id of software specification \n
                    **return type**: str\n

                **Example**


                 >>> asset_uid = client.software_specifications.get_uid(sw_spec_details)

        """
        SwSpec._validate_type(sw_spec_details, u'sw_spec_details', object, True)
        SwSpec._validate_type_of_details(sw_spec_details, SW_SPEC_DETAILS_TYPE)

        return WMLResource._get_required_element_from_dict(sw_spec_details, u'sw_spec_details',
                                                           [u'metadata', u'asset_id'])

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_id_by_name(self, sw_spec_name):
        """
                Get Unique Id of software specification.

                **Parameters**

                .. important::

                   #. **sw_spec_name**:  Name of the software specification\n
                      **type**: str\n

                **Output**

                .. important::

                    **returns**: Unique Id of software specification \n
                    **return type**: str\n

                **Example**


                 >>> asset_uid = client.software_specifications.get_id_by_name(sw_spec_name)

        """

        return SwSpec.get_uid_by_name(self, sw_spec_name)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_uid_by_name(self, sw_spec_name):
        """
                Get Unique Id of software specification. Deprecated!! Use get_id_by_name(self, sw_spec_name) instead

                **Parameters**

                .. important::

                   #. **sw_spec_name**:  Name of the software specification\n
                      **type**: str\n

                **Output**

                .. important::

                    **returns**: Unique Id of software specification \n
                    **return type**: str\n

                **Example**


                 >>> asset_uid = client.software_specifications.get_uid_by_name(sw_spec_name)

        """

        SwSpec._validate_type(sw_spec_name, u'sw_spec_uid', STR_TYPE, True)
        if not self._client.WSD_20:
            parameters = self._client._params(skip_space_project_chk=True)
            parameters.update(name=sw_spec_name)
            if not self._ICP:
                response = requests.get(self._client.service_instance._href_definitions.get_sw_specs_href(),
                                        params=parameters,
                                        headers=self._client._get_headers())
            else:

                response = requests.get(self._client.service_instance._href_definitions.get_sw_specs_href(),
                                        params=parameters,
                                        headers=self._client._get_headers(), verify=False)

            total_values = self._handle_response(200, u'list assets', response)["total_results"]
            if total_values != 0:
                sw_spec_details = self._handle_response(200, u'list assets', response)["resources"]
                return sw_spec_details[0][u'metadata'][u'asset_id']
            else:
                return "Not Found"
        else:
            return self.software_spec_list.get(sw_spec_name)


    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_href(sw_spec_details):
        """
            Get url of software specification.

           **Parameters**

           .. important::
                #. **sw_spec_details**:  software specification details\n
                   **type**: dict\n

           **Output**

           .. important::
                **returns**: href of software specification \n
                **return type**: str\n

           **Example**

             >>> sw_spec_details = client.software_specifications.get_details(sw_spec_uid)
             >>> sw_spec_href = client.software_specifications.get_href(sw_spec_details)

        """
        SwSpec._validate_type(sw_spec_details, u'sw_spec_details', object, True)
        SwSpec._validate_type_of_details(sw_spec_details, SW_SPEC_DETAILS_TYPE)

        return WMLResource._get_required_element_from_dict(sw_spec_details, u'sw_spec_details', [u'metadata', u'href'])


    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def delete(self, sw_spec_uid):
        """
            Delete a software specification.

            **Parameters**

            .. important::
                #. **sw_spec_uid**: Unique Id of software specification\n
                   **type**: str\n

            **Output**

            .. important::
                **returns**: status ("SUCCESS" or "FAILED")\n
                **return type**: str\n

            **Example**

             >>> client.software_specifications.delete(sw_spec_uid)

        """
        if self._client.WSD_20:
            raise WMLClientError(u'delete API is not supported in Watson Studio Desktop.')

        SwSpec._validate_type(sw_spec_uid, u'sw_spec_uid', STR_TYPE, True)

        if not self._ICP:
            response = requests.delete(self._client.service_instance._href_definitions.get_sw_spec_href(sw_spec_uid), params=self._client._params(),
                                    headers=self._client._get_headers())
        else:
            response = requests.delete(self._client.service_instance._href_definitions.get_sw_spec_href(sw_spec_uid), params=self._client._params(),
                                      headers=self._client._get_headers(), verify=False)
        if response.status_code == 200:
            return self._get_required_element_from_response(response.json())
        else:
            return self._handle_response(204, u'delete software specification', response)


    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def add_package_extension(self, sw_spec_uid, pkg_extn_id):
        """
                Add a package extension to software specifications existing metadata.

                **Parameters**

                .. important::

                    #. **sw_spec_uid**:  Unique Id of software specification which should be updated\n
                       **type**: str\n
                    #. **pkg_extn_id**:  Unique Id of package extension which should needs to added to software specification\n
                       **type**: str\n

                **Example**

                >>> client.software_specifications.add_package_extension(sw_spec_uid, pkg_extn_id)

        """
        WMLResource._chk_and_block_create_update_for_python36(self)
        if self._client.WSD_20:
            raise WMLClientError(u'package extension APIs are not supported in Watson Studio Desktop.')

        ##For CP4D, check if either spce or project ID is set
        self._client._check_if_either_is_set()

        sw_spec_uid = str_type_conv(sw_spec_uid)
        self._validate_type(sw_spec_uid, u'sw_spec_uid', STR_TYPE, True)
        self._validate_type(pkg_extn_id, u'pkg_extn_id', STR_TYPE, True)

        url = self._client.service_instance._href_definitions.get_sw_spec_href(sw_spec_uid)

        url = url + "/package_extensions/" + pkg_extn_id


        if not self._ICP:
            response = requests.put(url, params=self._client._params(), headers=self._client._get_headers())
        else:
            response = requests.put(url,  params=self._client._params(), headers=self._client._get_headers(), verify=False)

        if response.status_code == 204:
            print("SUCCESS")
            return "SUCCESS"
        else:
            return self._handle_response(204, u'pkg spec add', response, False)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def delete_package_extension(self, sw_spec_uid, pkg_extn_id):
        """
                Delete a package extension from software specifications existing metadata.

                **Parameters**

                .. important::

                    #. **sw_spec_uid**:  Unique Id of software specification which should be updated\n
                       **type**: str\n
                    #. **pkg_extn_id**:  Unique Id of package extension which should needs to deleted from software specification\n
                       **type**: str\n

                **Example**

                 >>> client.software_specifications.delete_package_extension(sw_spec_uid, pkg_extn_id)

        """
        if self._client.WSD_20:
            raise WMLClientError(u'package extension APIs are not supported in Watson Studio Desktop.')

        ##For CP4D, check if either spce or project ID is set
        self._client._check_if_either_is_set()

        sw_spec_uid = str_type_conv(sw_spec_uid)
        self._validate_type(sw_spec_uid, u'sw_spec_uid', STR_TYPE, True)
        self._validate_type(pkg_extn_id, u'pkg_extn_id', STR_TYPE, True)

        url = self._client.service_instance._href_definitions.get_sw_spec_href(sw_spec_uid)

        url = url + "/package_extensions/" + pkg_extn_id

        if not self._ICP:
            response = requests.delete(url,
                                       params=self._client._params(),
                                       headers=self._client._get_headers())
        else:
            response = requests.delete(url,
                                       params=self._client._params(),
                                       headers=self._client._get_headers(),
                                       verify=False)

        return self._handle_response(204, u'pkg spec delete', response, False)

    def _get_required_element_from_response(self, response_data):

        WMLResource._validate_type(response_data, u'sw_spec_response', dict)
        try:
            if self._client.default_space_id is not None:
                new_el = {'metadata': {
                                       'name': response_data['metadata']['name'],
                                       'asset_id': response_data['metadata']['asset_id'],
                                       'href': response_data['metadata']['href'],
                                       'asset_type': response_data['metadata']['asset_type'],
                                       'created_at': response_data['metadata']['created_at']
                                       #'updated_at': response_data['metadata']['updated_at']
                                       },
                          'entity': response_data['entity']

                          }
            elif self._client.default_project_id is not None:
                if self._client.WSD:

                    href = self._client.service_instance._href_definitions.get_base_asset_href(response_data['metadata']['asset_id']) + "?" + "project_id=" + response_data['metadata']['project_id']

                    new_el = {'metadata': {
                                           'name': response_data['metadata']['name'],
                                           'asset_id': response_data['metadata']['asset_id'],
                                           'href': response_data['metadata']['href'],
                                           'asset_type': response_data['metadata']['asset_type'],
                                           'created_at': response_data['metadata']['created_at']
                                           },
                              'entity': response_data['entity']

                              }
                else:
                    new_el = {'metadata': {
                                           'name': response_data['metadata']['name'],
                                           'asset_id': response_data['metadata']['asset_id'],
                                           'href': response_data['metadata']['href'],
                                           'asset_type': response_data['metadata']['asset_type'],
                                           'created_at': response_data['metadata']['created_at']
                                       },
                             'entity': response_data['entity']

                            }

            else:
                # For system software spec
                new_el = {'metadata': {
                                       'name': response_data['metadata']['name'],
                                       'asset_id': response_data['metadata']['asset_id'],
                                       'href': response_data['metadata']['href'],
                                       'asset_type': response_data['metadata']['asset_type'],
                                       'created_at': response_data['metadata']['created_at']
                                       #'updated_at': response_data['metadata']['updated_at']
                                       },
                          'entity': response_data['entity']

                          }
            if (self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES) and 'href' in response_data['metadata']:
                href_without_host = response_data['metadata']['href'].split('.com')[-1]
                new_el[u'metadata'].update({'href': href_without_host})
            return new_el
        except Exception as e:
            raise WMLClientError("Failed to read Response from down-stream service: " + response_data.text)



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
import os
import time
import datetime
from configparser import ConfigParser
from functools import wraps
from time import sleep
import ibm_boto3

from ibm_watson_machine_learning.tests.utils.cleanup import bucket_cleanup

__all__ = [
    "get_wml_credentials",
    "get_cos_credentials",
    "get_db_credentials",
    "get_env",
    "is_cp4d",
    "bucket_exists",
    "bucket_name_gen",
    "create_bucket",
    'print_test_separators',
    "get_space_id",
    "resources_discovery",
    "find_not_freed_resources",
    "setup_nfs_env",
    "save_data_to_container",
    "save_data_to_cos_bucket",
    "create_connection_to_cos"
]

if "ENV" in os.environ:
    environment = os.environ['ENV']
else:
    environment = "YP_QA"

timeouts = "TIMEOUTS"
credentials = "CREDENTIALS"
training_data = "TRAINING_DATA"
configDir = "./config.ini"

config = ConfigParser()
config.read(configDir)


def get_env():
    return environment


def get_wml_credentials(env=environment):
    return json.loads(config.get(env, 'wml_credentials'))


def get_cos_credentials(env=environment):
    return json.loads(config.get(env, 'cos_credentials'))


def get_db_credentials(db_name, env=environment):
    """
    Returns credentials for a given database.
    Credentials should be added to config.ini
    with the following template: `<database_name>_credentials: {...}`
    """
    return json.loads(config.get(env, f"{db_name}_credentials"))


def is_cp4d():
    if "CP4D" in get_env():
        return True
    elif "ICP" in get_env():
        return True
    elif "OPEN_SHIFT" in get_env():
        return True
    elif "CPD" in get_env():
        return True

    return False


def bucket_exists(cos_resource, bucket_name):
    """
    Return True if bucket with `bucket_name` exists. Else False.
    """
    buckets = cos_resource.buckets.all()
    for bucket in buckets:
        if bucket.name == bucket_name:
            return True
    print("Bucket {0} not found".format(bucket_name))
    return False


def bucket_name_gen(prefix='bucket-tests', id_size=8):
    import random
    import string

    return prefix + "-" + ''.join(random.choice(string.ascii_lowercase + string.digits) for x in range(id_size))


def create_bucket(cos_resource, bucket_name):
    """
    Creates bucket of with a given name and timestamp suffix.
    """
    try:
        bucket_cleanup(cos_resource, prefix=f"{bucket_name}-")
    except Exception as e:
        print(f"Bucket cleanup with prefix {bucket_name}- failed due to:\n{e}\n skipped")

    bucket_name = bucket_name_gen(prefix=f"{bucket_name}-{str(datetime.date.today())}")
    print(f"Creating COS bucket: {bucket_name}")
    cos_resource.Bucket(bucket_name).create()

    return bucket_name


def print_test_separators(method):
    """Printing visual separators for tests."""

    @wraps(method)
    def _method(*method_args, **method_kwargs):
        print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
        method_output = method(*method_args, **method_kwargs)
        print("________________________________________________________________________")

        return method_output

    return _method


def get_space_id(wml_client, space_name, cos_resource_instance_id=None):
    """
    Return space id of existed or just created space named as `space_name`.
    """
    spaces_details = wml_client.spaces.get_details(limit=200).get('resources')
    space_id = None
    for space_d in spaces_details:
        if space_d['entity']['name'] == space_name:
            space_id = space_d['metadata']['id']

    if not space_id:
        # create new space for tests
        if wml_client.CLOUD_PLATFORM_SPACES:
            metadata = {
                wml_client.spaces.ConfigurationMetaNames.NAME: space_name,
                wml_client.spaces.ConfigurationMetaNames.STORAGE: {
                    'type': 'bmcos_object_storage',
                    "resource_crn": cos_resource_instance_id},
                wml_client.spaces.ConfigurationMetaNames.COMPUTE: {
                    "name": get_wml_credentials()['name'],
                    "crn": get_wml_credentials()['iam_serviceid_crn']}
            }
        else:
            metadata = {
                wml_client.spaces.ConfigurationMetaNames.NAME: space_name
            }
        details = wml_client.spaces.store(meta_props=metadata, background_mode=False)
        space_id = details['metadata'].get('id')
        print(f"New space `{space_name}` has been created, space_id={space_id}")
        sleep(5)  # wait for space preparing

    return space_id


def resources_discovery(wml_client):
    resources_map = {}

    def check(parent_el, parent_name):
        for r in parent_el.__dict__:
            try:
                if parent_el.__dict__[r].__class__.__module__.startswith('ibm_watson_machine_learning.'):
                    details = parent_el.__dict__[r].get_details()['resources']
                    if type(details) is list:
                        try:
                            uids = [x['metadata']['id'] for x in details]
                        except:
                            uids = [x['metadata']['guid'] for x in details]
                        print(f"{parent_name}.{r}: {len(uids)}")
                        resources_map[f"{parent_name}.{r}"] = uids

                        check(parent_el.__dict__[r], f"{parent_name}.{r}")
            except Exception as e:
                pass

    check(wml_client, "wml_client")

    return resources_map


def find_not_freed_resources(initial_list, final_list):
    not_freed_list = {}
    for k in initial_list:
        initial_uids = initial_list[k]
        final_uids = final_list[k]
        not_freed_uids = [uid for uid in final_uids if uid not in initial_uids]
        if len(not_freed_uids) > 0:
            not_freed_list[k] = not_freed_uids

    if len(not_freed_list) > 0:
        raise Exception(f"Resources not freed: {not_freed_list}")


def setup_nfs_env(wml_credentials, dataset_file_name, remote_directory='datasets'):
    import paramiko
    from scp import SCPClient

    ssh_host = wml_credentials.get('ssh_host')
    ssh_username = wml_credentials.get('ssh_username')
    ssh_password = wml_credentials.get('ssh_password')

    kube_username = wml_credentials.get('kube_username')
    kube_password = None

    if not (ssh_host and ssh_username and ssh_password):
        print("NFS setup omitted. Required credentials missing: ssh_host, ssh_username, ssh_password")
        return

    host = wml_credentials['url']
    username = wml_credentials['username']
    password = wml_credentials['password']

    storage_config_filepath = "autoai-sdk-test-storage-class-config.yaml"
    pvc_config_filepath = "autoai-sdk-test-pvc-config.yaml"
    storage_config_file_name = storage_config_filepath.split('/')[-1]
    pvc_config_file_name = pvc_config_filepath.split('/')[-1]
    storage_class_name = "managed-nfs-storage"
    pvc_name = "autoai-sdk-test-pvc"
    volume_name = "autoai-sdk-test-volume"

    project_id = wml_credentials['project_id']
    connection_id = None
    namespace = host.replace('https://', '').split('-')[0]

    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    ssh.connect(ssh_host, username=ssh_username, password=ssh_password)

    def run_ssh(cmd, format='plain'):
        ssh_stdin, ssh_stdout, ssh_stderr = ssh.exec_command(cmd)
        out = ssh_stdout.read().decode('utf-8')
        err = ssh_stderr.read().decode('utf-8')

        if err and not ('200 OK' in err or '409 Conflict' in err or '201 Created' in err):
            raise Exception(err)

        if format == 'json':
            return json.loads(out.strip().split('\n')[-1])
        elif format == 'plain':
            return out
        else:
            raise Exception(f'Unsupported return value format: {format}')

    def put_file(filepath, directory='.'):
        scp = SCPClient(ssh.get_transport())
        scp.put(filepath, recursive=True, remote_path=f'~/{directory}')
        scp.close()

    kube_password = run_ssh('cat /root/auth/kubeadmin-password')

    res = run_ssh(f'oc login -u {kube_username} -p {kube_password} --kubeconfig /root/auth/kubeconfig')

    try:
        res = run_ssh(f'oc project {namespace} --kubeconfig /root/auth/kubeconfig')
        print(res)
    except:
        raise Exception(
            "Cannot change project to correct one for this cluster. Probably cluster changed. Add to credentials kube_username and make sure ssh credentials match the cluster.")

    # Check if there is already an existing NFS storage class
    res = run_ssh('oc get storageclass --kubeconfig /root/auth/kubeconfig')

    # Create a PVC volume
    if storage_class_name in res:
        print(f"Storage class exists: {storage_class_name}")
    else:
        res = run_ssh('ls')
        if storage_config_file_name not in res:
            put_file(storage_config_filepath)

        res = run_ssh(f'oc create -f ./{storage_config_file_name} --kubeconfig /root/auth/kubeconfig')

        if not ' created' in res:
            raise Exception(f'Failure during creating storage class: {res}')
        else:
            print(f"Storage class created: {res.replace(' created', '')}")

    res = run_ssh('oc get pvc --kubeconfig /root/auth/kubeconfig')

    if pvc_name in res:
        print(f"PVC exists: {pvc_name}")
    else:
        res = run_ssh('ls')
        if pvc_config_file_name not in res:
            put_file(pvc_config_filepath)

        res = run_ssh(f'oc create -f ./{pvc_config_file_name} --kubeconfig /root/auth/kubeconfig')

        if not ' created' in res:
            raise Exception(f'Failure during creating PVC: {res}')
        else:
            print(f"PVC created: {res.replace(' created', '')}")

    # Create a Volume on your existing PVC

    res = run_ssh(f"""
        curl -kiv -X POST \
            {host}/icp4d-api/v1/authorize \
            -H 'Content-Type: application/json' \
            -d '{{
                    "username":"{username}",
                    "password":"{password}"
            }}'
    """, format='json')

    token = res['token']

    print(f"Successfully got token")

    res = run_ssh(f"""
            curl -kiv -X GET \
                -H 'Authorization: Bearer {token}' \
                {host}/zen-data/v3/service_instances?addon_type=volumes
    """, format='json')

    instance = None
    for e in res['service_instances']:
        if e['display_name'] == volume_name and e['metadata']['zenControlPlaneNamespace'] == namespace \
                and e['metadata']['existing_pvc_name'] == pvc_name:
            instance = e
            break

    if instance:
        print(f"Service instance already exists: {instance['id']}")
    else:
        print("No service instance found, creating one.")

        res = run_ssh(f"""
            curl -kiv -X POST \
                -H 'Authorization: Bearer {token}' \
                -H 'Content-Type: application/json' \
                -d '{{
                    "addon_type": "volumes",
                    "addon_version": "-",
                    "create_arguments": {{
                        "metadata": {{
                            "existing_pvc_name": "{pvc_name}"
                        }}
                    }},
                    "namespace": "{namespace}",
                    "display_name":"{volume_name}"
                }}' \
                {host}/zen-data/v3/service_instances
        """, format='json')

        instance = res
        print(f"Service instance created: {instance['id']}")

    # Start the volume

    res = run_ssh(f"""
        curl -kiv --location --request POST '{host}/zen-data/v1/volumes/volume_services/{volume_name}' \
            --header 'Authorization: Bearer {token}' \
            --header 'Content-Type: application/json' \
            --data-raw '{{}}'
    """, format='json')

    started_earlier = False

    if '_statusCode_' in res and res['_statusCode_'] == 409:
        print(f'The volume `{volume_name}` is already started')
        started_earlier = True
    else:
        print(f'Successfully started volume: {volume_name}')

    for i in range(100):
        res = run_ssh(f"""
            curl -kiv --location --request GET '{host}/zen-data/v3/service_instances/{instance['id']}/?include_service_status=true' \
                --header 'Authorization: Bearer {token}'
        """, format='json')

        time.sleep(2)

        if res['services_status'] == 'RUNNING':
            print(f'Volume `{volume_name}` is started and running.')
            break

    if res['services_status'] == 'FAILED':
        raise Exception(
            f'Volume `{volume_name}` failed to run. Probably wrong cluster is set in login and as project. The project when `oc project` should be `{namespace}`')

    if res['services_status'] != 'RUNNING':
        raise Exception(f'Volume `{volume_name}` not running.')

    # Upload a file into the volume
    directory_exists = True

    try:
        res = run_ssh(f"""
            curl -kiv --location --request GET '{host}/zen-volumes/{volume_name}/v1/volumes/directories/{remote_directory}' \
                --header 'Authorization: Bearer {token}'
        """, format='json')
    except Exception as e:
        if '404 Not Found' in str(e):
            directory_exists = False
        else:
            raise e

    if directory_exists and dataset_file_name in res['responseObject']['directoryContents']:
        print('File is already uploaded.')
    else:
        res = run_ssh(f'ls')

        if remote_directory not in res.split():
            res = run_ssh(f'mkdir {remote_directory}')

        res = run_ssh(f'ls {remote_directory}')
        if not dataset_file_name in res:
            put_file(f"autoai/data/{dataset_file_name}", remote_directory)

        res = run_ssh(f"""
            curl -kiv --location --request PUT '{host}/zen-volumes/{volume_name}/v1/volumes/files/{remote_directory}%2F{dataset_file_name}' \
                --header 'Authorization: Bearer {token}' \
                --form 'upFile=@./{remote_directory}/{dataset_file_name}'
        """, format='json')

        print(res)

    # Create the connection

    res = run_ssh(f"""
        curl -kiv --location --request GET '{host}/v2/connections?project_id={project_id}' \
            --header 'Authorization: Bearer {token}'
    """, format='json')

    for e in res['resources']:
        if e['entity']['properties'].get('volume') == volume_name:
            connection_id = e['metadata']['asset_id']
            break

    if connection_id:
        print(f'Connection already exists: {connection_id}')
    else:
        res = run_ssh(f"""
            curl -kiv --location --request POST '{host}/v2/connections?project_id={project_id}' \
                --header 'Authorization: Bearer {token}' \
                --header 'Content-Type: application/json' \
                --data-raw '{{ "datasource_type": "9f30e3c3-b854-4144-b5c3-98b7a835dc79", "owner_id": "1000330999", "name": "{remote_directory}", "origin_country" : "us", "properties": {{ "volume": "{volume_name}", "trust_all_ssl_cert": "false", "password": "{password}", "gateway_url": "https://internal-nginx-svc:12443", "username": "{username}" }} }}'
        """, format='json')

        connection_id = res['metadata']['asset_id']

        print(f'Connection created: {connection_id}')

    # Get asset id
    # TODO

    asset_id = None

    ssh.close()

    return connection_id, f"/{remote_directory}/{dataset_file_name}", asset_id


def save_data_to_container(data_path, data_cos_path, wml_client, data_filenames=None):
    import requests

    if wml_client.default_space_id:
        details = wml_client.spaces.get_details(wml_client.default_space_id)

    elif wml_client.default_project_id:
        project_url = wml_client.service_instance._href_definitions.get_project_href(
            wml_client.default_project_id)
        if wml_client.ICP:
            response = requests.get(
                project_url,
                headers=wml_client._get_headers(),
                verify=False
            )

        else:
            response = requests.get(
                project_url,
                headers=wml_client._get_headers()
            )

        details = response.json()
    else:
        raise ValueError("WML client should have set default project or space id.")

    properties = details['entity']['storage']['properties']
    properties.update(details['entity']['storage']['properties']['credentials']['editor'])
    properties.update(details['entity']['storage']['properties']['credentials']['editor'])

    save_data_to_cos_bucket(data_path=data_path,
                            data_cos_path=data_cos_path,
                            access_key_id=properties.get('access_key_id'),
                            secret_access_key=properties.get('secret_access_key'),
                            cos_endpoint=properties['endpoint_url'],
                            bucket_name=properties['bucket_name'],
                            data_filenames=data_filenames,
                            is_container=True)


def save_data_to_cos_bucket(data_path, data_cos_path, access_key_id, secret_access_key, cos_endpoint, bucket_name,
                            data_filenames=None, is_container=False):
    cos_resource = ibm_boto3.resource(
        service_name="s3",
        endpoint_url=cos_endpoint,
        aws_access_key_id=access_key_id,
        aws_secret_access_key=secret_access_key
    )

    # Prepare bucket
    if not is_container:
        if not bucket_exists(cos_resource, bucket_name):
            bucket_name = create_bucket(cos_resource, bucket_name)

            assert bucket_exists(cos_resource, bucket_name)

    # Write data to cos bucket

    if os.path.isfile(data_path):
        cos_resource.Bucket(bucket_name).upload_file(data_path, data_cos_path)
    elif os.path.isdir(data_path):
        # upload multiple files
        for file in data_filenames:
            data_file_path = os.path.join(data_path, file) if data_path else file
            cos_file_path = os.path.join(data_cos_path, file) if data_cos_path else file

            cos_resource.Bucket(bucket_name).upload_file(data_file_path, cos_file_path)

    else:
        raise ValueError("data_path is not valid directory or file.")

    return bucket_name


def create_connection_to_cos(wml_client, cos_credentials, cos_endpoint, bucket_name, save_data=False, data_path=None,
                             data_cos_path=None, data_filenames=None):
    if save_data and data_path and data_cos_path:
        bucket_name = save_data_to_cos_bucket(data_path, data_cos_path,
                                              cos_credentials['cos_hmac_keys']['access_key_id'],
                                              cos_credentials['cos_hmac_keys']['secret_access_key'], cos_endpoint,
                                              bucket_name,
                                              data_filenames=data_filenames,
                                              is_container=False)

    auth_endpoint = cos_credentials.get('auth_endpoint')
    if auth_endpoint is None:
        auth_endpoint = wml_client.service_instance._href_definitions.get_iam_token_url()

    connection_details = wml_client.connections.create({
        'datasource_type': wml_client.connections.get_datasource_type_uid_by_name('bluemixcloudobjectstorage'),
        'name': f'Connection to COS for tests {datetime.datetime.utcnow().isoformat()}',
        'properties': {
            'bucket': bucket_name,
            'access_key': cos_credentials['cos_hmac_keys']['access_key_id'],
            'secret_key': cos_credentials['cos_hmac_keys']['secret_access_key'],
            'iam_url': auth_endpoint,
            'url': cos_endpoint
        }
    })

    connection_id = wml_client.connections.get_uid(connection_details)

    return connection_id, bucket_name



"""
Module to where data handler are implemented.
"""
#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import abc

from ibmfl.data.data_handler import DataHandler
from ibmfl.data.env_spec import EnvHandler


class EnvDataHandler(DataHandler):
    """
    Base class to load data and  environment for reinforcement learning.
    """

    @abc.abstractmethod
    def get_data(self, **kwargs):
        """
        Read train data and test data for reinforcement learning
        :return:
        """

    @abc.abstractmethod
    def get_env_class_ref(self) -> EnvHandler:
        """
           Get environment reference for RL trainer, the instance is created in
           model class as part of trainer initialization
        """



"""
 An enumeration class for the message type field which describe what
 kind of data is being sent inside the Message
"""
#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from enum import Enum

__author__ = "Supriyo Chakraborty, Shalisha Witherspoon, Dean Steuer"


class MessageType(Enum):
    """
    Message types for communication between party and aggregator
    """
    MODEL_UPDATE = 1
    MODEL_HYPERPARAMETERS = 2
    MODEL_PARAMETERS = 3
    REQUEST_MODEL_HYPERPARAMETERS = 4
    REQUEST_MODEL_UPDATE = 5
    REGISTER = 6
    TRAIN = 7
    SAVE_MODEL = 8
    PREDICT_MODEL = 9
    EVAL_MODEL = 10
    ACK = 11
    SYNC_MODEL = 12
    STOP = 14
    ERROR_AUTH = 400



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest
import  os
import ibm_boto3
from ibm_watson_machine_learning.preprocessing import DataJoinGraph
from ibm_watson_machine_learning.helpers.connections import DataConnection, ContainerLocation, FSLocation
from ibm_watson_machine_learning.tests.utils import is_cp4d, save_data_to_container, create_connection_to_cos
from ibm_watson_machine_learning.tests.autoai.abstract_tests_classes import AbstractTestOBM, \
    AbstractTestBatch

from ibm_watson_machine_learning.utils.autoai.errors import WrongDataJoinGraphNodeName

from ibm_watson_machine_learning.utils.autoai.enums import PredictionType, Metrics, RegressionAlgorithms


class TestAutoAIRemote(AbstractTestOBM, unittest.TestCase):
    """
    The test can be run on CLOUD, and CPD
    """
    HISTORICAL_RUNS_CHECK = False
    cos_resource = None

    input_data_filenames = ["go_1k.csv", "go_daily_sales.csv", "go_methods.csv",
                            "go_products.csv", "go_retailers.csv"]

    input_data_path = './autoai/data/go_sales_1k/'

    input_node_names = [name.replace('go_', 'node_').split('.')[0] for name in input_data_filenames]

    cos_endpoint = "https://s3.us.cloud-object-storage.appdomain.cloud"
    data_cos_path = "data/"

    SPACE_ONLY = False

    OPTIMIZER_NAME = "Go Sales OBM test sdk"

    target_space_id: str = None

    experiment_info = dict(
        name=OPTIMIZER_NAME,
        desc='test description',
        prediction_type=PredictionType.REGRESSION,
        prediction_column='Quantity',
        scoring=Metrics.ROOT_MEAN_SQUARED_ERROR,
        # include_only_estimators=[RegressionAlgorithms.SnapDT, RegressionAlgorithms.RIDGE]
    )

    def test_00b_prepare_COS_instance_and_connection(self):
        TestAutoAIRemote.connection_id, TestAutoAIRemote.bucket_name = create_connection_to_cos(
            wml_client=self.wml_client,
            cos_credentials=self.cos_credentials,
            cos_endpoint=self.cos_endpoint,
            bucket_name=self.bucket_name,
            save_data=True,
            data_path=self.input_data_path,
            data_filenames=self.input_data_filenames,
            data_cos_path=self.data_cos_path)

        self.assertIsInstance(self.connection_id, str)

    def test_01_create_multiple_data_connections__connections_created(self):
        TestAutoAIRemote.data_connections = []
        for file, node_name in zip(self.input_data_filenames, self.input_node_names):
            asset_details = self.wml_client.data_assets.store({
                self.wml_client.data_assets.ConfigurationMetaNames.CONNECTION_ID: self.connection_id,
                self.wml_client.data_assets.ConfigurationMetaNames.NAME: f"{node_name} - training asset",
                self.wml_client.data_assets.ConfigurationMetaNames.DATA_CONTENT_NAME:
                    os.path.join(self.bucket_name, self.data_cos_path, file)
            })

            asset_id = self.wml_client.data_assets.get_id(asset_details)
            self.assertIsInstance(asset_id, str)

            conn = DataConnection(
                data_join_node_name=node_name,
                data_asset_id=asset_id
            )
            TestAutoAIRemote.data_connections.append(conn)

        TestAutoAIRemote.results_connection = None

        self.assertIsNotNone(obj=TestAutoAIRemote.data_connections)
        self.assertGreater(len(TestAutoAIRemote.data_connections), 0)
        self.assertIsNone(self.results_connection)


    def test_02_create_data_join_graph__graph_created(self):
        data_join_graph = DataJoinGraph()
        data_join_graph.node(name="node_daily_sales")
        data_join_graph.node(name="node_methods")
        data_join_graph.node(name="node_retailers")
        data_join_graph.node(name="node_products")
        data_join_graph.node(name="node_1k", main=True)

        data_join_graph.edge(from_node="node_products", to_node="node_1k",
                             from_column=["Product number"], to_column=["Product number"])
        data_join_graph.edge(from_node="node_retailers", to_node="node_1k",
                             from_column=["Retailer code"], to_column=["Retailer code"])
        data_join_graph.edge(from_node="node_methods", to_node="node_daily_sales",
                             from_column=["Order method code"], to_column=["Order method code"])

        TestAutoAIRemote.data_join_graph = data_join_graph

        print(f"data_join_graph: {data_join_graph}")




if __name__ == '__main__':
    unittest.main()



"""
Serialization factory provides a way to create a serializer
and deserializer to convert byte streams to a message and
vice versa
"""
#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ibmfl.message.json_serializer import JSONSerializer
from ibmfl.message.pickle_serializer import PickleSerializer
from ibmfl.message.serializer_types import SerializerTypes


class SerializerFactory(object):
    """
    Class for a factory to serialize and deserialize
    """
    def __init__(self, serializer_type):
        """
        Creates an object of `SerializerFactory` class

        :param serializer_type: type of seriaze and deserialize
        :type serializer_type: `Enum`
        """
        self.serializer = None
        if serializer_type is SerializerTypes.PICKLE:
            self.serializer = PickleSerializer()
        elif serializer_type is SerializerTypes.JSON_PICKLE:
            self.serializer = JSONSerializer()

    def build(self):
        """
        Returns a serializer

        :param: None
        :return: serializer
        :rtype: `Serializer`
        """
        return self.serializer



# coding: utf-8

"""
Copyright 2016 SmartBear Software

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Ref: https://github.com/swagger-api/swagger-codegen
"""

#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from pprint import pformat
from six import iteritems


class PatchOperationFunctions(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """
    def __init__(self, op=None, path=None, value=None, _from=None):
        """
        PatchOperationFunctions - a model defined in Swagger

        :param dict swaggerTypes: The key is attribute name
                                  and the value is attribute type.
        :param dict attributeMap: The key is attribute name
                                  and the value is json key in definition.
        """
        self.swagger_types = {
            'op': 'str',
            'path': 'str',
            'value': 'object',
            '_from': 'str'
        }

        self.attribute_map = {
            'op': 'op',
            'path': 'path',
            'value': 'value',
            '_from': 'from'
        }

        self._op = op
        self._path = path
        self._value = value
        self.__from = _from

    @property
    def op(self):
        """
        Gets the op of this PatchOperationFunctions.
        operation

        :return: The op of this PatchOperationFunctions.
        :rtype: str
        """
        return self._op

    @op.setter
    def op(self, op):
        """
        Sets the op of this PatchOperationFunctions.
        operation

        :param op: The op of this PatchOperationFunctions.
        :type: str
        """
        allowed_values = ["add", "remove", "replace", "move", "copy", "test"]
        if op not in allowed_values:
            raise ValueError(
                "Invalid value for `op`, must be one of {0}"
                .format(allowed_values)
            )
        self._op = op

    @property
    def path(self):
        """
        Gets the path of this PatchOperationFunctions.
        path

        :return: The path of this PatchOperationFunctions.
        :rtype: str
        """
        return self._path

    @path.setter
    def path(self, path):
        """
        Sets the path of this PatchOperationFunctions.
        path

        :param path: The path of this PatchOperationFunctions.
        :type: str
        """
        self._path = path

    @property
    def value(self):
        """
        Gets the value of this PatchOperationFunctions.


        :return: The value of this PatchOperationFunctions.
        :rtype: object
        """
        return self._value

    @value.setter
    def value(self, value):
        """
        Sets the value of this PatchOperationFunctions.


        :param value: The value of this PatchOperationFunctions.
        :type: object
        """
        self._value = value

    @property
    def _from(self):
        """
        Gets the _from of this PatchOperationFunctions.


        :return: The _from of this PatchOperationFunctions.
        :rtype: str
        """
        return self.__from

    @_from.setter
    def _from(self, _from):
        """
        Sets the _from of this PatchOperationFunctions.


        :param _from: The _from of this PatchOperationFunctions.
        :type: str
        """
        self.__from = _from

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other




#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest
import json
from configparser import ConfigParser

from ibm_watson_machine_learning.experiment.autoai.optimizers import RemoteAutoPipelines
from ibm_watson_machine_learning.helpers.connections import DataConnection, DatabaseLocation
from ibm_watson_machine_learning.tests.utils import is_cp4d, get_env
from ibm_watson_machine_learning.tests.autoai.abstract_tests_classes import (
    AbstractTestAutoAIRemote)

from ibm_watson_machine_learning.wml_client_error import WMLClientError

configDir = "./config.ini"

config = ConfigParser()
config.read(configDir)


@unittest.skip("Skip due to issue 25182")
@unittest.skipIf(not is_cp4d(), "JDBC is not tested on cloud yet.")
class TestAutoAIRemote(AbstractTestAutoAIRemote, unittest.TestCase):
    """
    The test can be run on CPD
    The test covers:
    - JDBC DB2 connection set-up
    - downloading training data from connection
    - downloading all generated pipelines to lale pipeline
    - deployment with lale pipeline
    - deployment deletion
    Connection used in test:
     - input: Database connection pointing to DB2.
     - output: null
    """
    prediction_column = 'SPECIES'
    connection_asset_details = json.loads(config.get(get_env(), 'jdbc_db2_connection_asset_details'))

    def test_00c_prepare_connection_to_db2(self):
        # import urllib
        # import requests
        #
        # driver_file_path = self.connection_asset_details['driver_file_path']
        # driver_file_name = driver_file_path.split('/')[-1]
        #
        # signed_url = self.wml_client.service_instance._href_definitions.get_wsd_model_attachment_href() + '/' + \
        #                              urllib.parse.quote("dbdrivers/" + driver_file_name + "/sign", safe='')
        # params = self.wml_client._params()
        #
        # params['expires_in'] = 60*60*24
        #
        # response = requests.post(
        #     signed_url,
        #     headers=self.wml_client._get_headers(no_content_type=True),
        #     params=params,
        #     verify=False
        # )
        #
        # print(response.json())


        # with open(driver_file_path, 'rb') as fdata:
        #     content_upload_url = self.wml_client.service_instance._href_definitions.get_wsd_model_attachment_href() + '/' + \
        #                          urllib.parse.quote("dbdrivers/" + driver_file_name, safe='')
        #     response = requests.put(
        #         content_upload_url,
        #         files={'file': ('native', fdata, 'application/octet-stream', {'Expires': '0'})},
        #         headers=self.wml_client._get_headers(no_content_type=True),
        #         params=self.wml_client._params(),
        #         verify=False
        #     )
        #     if response.status_code == 201:
        #         print(response.text)
        #         # asset_body = {
        #         #     "asset_type": "wml_model",
        #         #     "name": "native",
        #         #     "object_key": asset_uid + "/" + file_name_to_attach,
        #         #     "object_key_is_read_only": True
        #         # }
        #         # attach_payload = json.dumps(asset_body, separators=(',', ':'))
        #         #
        #         # attach_response = requests.post(attach_url,
        #         #                                 data=attach_payload,
        #         #                                 params=params,
        #         #                                 headers=headers,
        #         #                                 verify=False)
        #     else:
        #         print(response.text)
        #         raise Exception(f"Failed while uploading driver: {response.text}")

        connection_details = self.wml_client.connections.create({
            'datasource_type': self.wml_client.connections.get_datasource_type_uid_by_name(
                self.connection_asset_details['datasource_type_name']
            ),
            'name': 'Connection to DB2 for tests',
            'properties': self.connection_asset_details['properties']
        })

        TestAutoAIRemote.connection_id = self.wml_client.connections.get_uid(connection_details)
        self.assertIsInstance(self.connection_id, str)

    def test_00d_prepare_connected_data_asset(self):
        asset_details = self.wml_client.data_assets.store({
            self.wml_client.data_assets.ConfigurationMetaNames.CONNECTION_ID: self.connection_id,
            self.wml_client.data_assets.ConfigurationMetaNames.NAME: "Iris - training asset",
            self.wml_client.data_assets.ConfigurationMetaNames.DATA_CONTENT_NAME: f"/{self.connection_asset_details['schema_name']}/{self.connection_asset_details['table_name']}",
            "connectionPath": f"/{self.connection_asset_details['schema_name']}/{self.connection_asset_details['table_name']}"
        })

        print(asset_details)

        TestAutoAIRemote.asset_id = self.wml_client.data_assets.get_id(asset_details)
        self.assertIsInstance(self.asset_id, str)

    def test_02_DataConnection_setup(self):
        TestAutoAIRemote.data_connection = DataConnection(data_asset_id=self.asset_id)

        TestAutoAIRemote.results_connection = None

        self.assertIsNotNone(obj=TestAutoAIRemote.data_connection)
        # self.assertIsNotNone(obj=TestAutoAIRemote.results_connection)

    def test_02a_read_saved_remote_data_before_fit(self):
        TestAutoAIRemote.data = self.data_connection.read()
        print("Data sample:")
        print(self.data.head())
        self.assertGreater(len(self.data), 0)

    def test_03_initialize_optimizer(self):
        AbstractTestAutoAIRemote.remote_auto_pipelines = self.experiment.optimizer(
            name='Iris - AutoAI',
            prediction_type=self.experiment.PredictionType.MULTICLASS,
            prediction_column=self.prediction_column,
            scoring=self.experiment.Metrics.ACCURACY_SCORE,
            drop_duplicates=False,
            enable_all_data_sources=True,

        )

        self.assertIsInstance(self.remote_auto_pipelines, RemoteAutoPipelines,
                              msg="experiment.optimizer did not return RemoteAutoPipelines object")

    def test_29_delete_connection_and_connected_data_asset(self):
        self.wml_client.data_assets.delete(self.asset_id)
        self.wml_client.connections.delete(self.connection_id)

        with self.assertRaises(WMLClientError):
            self.wml_client.connections.get_details(self.connection_id)


if __name__ == '__main__':
    unittest.main()



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest
import time
import pandas as pd
import logging
from ibm_watson_machine_learning.wml_client_error import ApiRequestFailure
from ibm_watson_machine_learning.tests.ICP.preparation_and_cleaning import *


class TestWMLClientWithDO(unittest.TestCase):
    deployment_uid = None
    model_uid = None
    scoring_url = None
    scoring_uid = None
    logger = logging.getLogger(__name__)
    skip_tests = False

    @classmethod
    def setUpClass(self):
        TestWMLClientWithDO.logger.info("Service Instance: setting up credentials")
        self.wml_credentials = get_wml_credentials()
        self.client = get_client()
        self.model_path = os.path.join(os.getcwd(), 'artifacts', 'do-model.tar.gz')

        if self.wml_credentials['version'] == '2.0':
           TestWMLClientWithDO.skip_tests = True

    def test_01_set_space(self):
        if not TestWMLClientWithDO.skip_tests:
            space = self.client.spaces.store({self.client.spaces.ConfigurationMetaNames.NAME: "DO_test_case"})

            TestWMLClientWithDO.space_id = self.client.spaces.get_uid(space)
            self.client.set.default_space(TestWMLClientWithDO.space_id)
            self.assertTrue("SUCCESS" in self.client.set.default_space(TestWMLClientWithDO.space_id))


    # Doesn't work since it doesn't recognize the class variables
    # @unittest.skipUnless(not skip_tests, 'WML Server 2.0')
    def test_02_publish_do_model_in_repository(self):
        if not TestWMLClientWithDO.skip_tests:
            TestWMLClientWithDO.logger.info("Saving trained model in repo ...")
            TestWMLClientWithDO.logger.debug("Model path: {}".format(self.model_path))

            self.client.repository.ModelMetaNames.show()

            sw_spec_uid = self.client.software_specifications.get_uid_by_name("do_12.9")
            output_data_schema = [{'id': 'stest',
                                  'type': 'list',
                                  'fields': [{'name': 'age', 'type': 'float'},
                                             {'name': 'sex', 'type': 'float'},
                                             {'name': 'cp', 'type': 'float'},
                                             {'name': 'restbp', 'type': 'float'},
                                             {'name': 'chol', 'type': 'float'},
                                             {'name': 'fbs', 'type': 'float'},
                                             {'name': 'restecg', 'type': 'float'},
                                             {'name': 'thalach', 'type': 'float'},
                                             {'name': 'exang', 'type': 'float'},
                                             {'name': 'oldpeak', 'type': 'float'},
                                             {'name': 'slope', 'type': 'float'},
                                             {'name': 'ca', 'type': 'float'},
                                             {'name': 'thal', 'type': 'float'}]
                                   }, {'id': 'teste2',
                     'type': 'test',
                     'fields': [{'name': 'age', 'type': 'float'},
                                {'name': 'sex', 'type': 'float'},
                                {'name': 'cp', 'type': 'float'},
                                {'name': 'restbp', 'type': 'float'},
                                {'name': 'chol', 'type': 'float'},
                                {'name': 'fbs', 'type': 'float'},
                                {'name': 'restecg', 'type': 'float'},
                                {'name': 'thalach', 'type': 'float'},
                                {'name': 'exang', 'type': 'float'},
                                {'name': 'oldpeak', 'type': 'float'},
                                {'name': 'slope', 'type': 'float'},
                                {'name': 'ca', 'type': 'float'},
                                {'name': 'thal', 'type': 'float'}]}]
            model_meta_props = {self.client.repository.ModelMetaNames.NAME: "LOCALLY created DO model",
                           self.client.repository.ModelMetaNames.TYPE: "do-docplex_12.9",
                           self.client.repository.ModelMetaNames.SOFTWARE_SPEC_UID: sw_spec_uid,
                           self.client.repository.ModelMetaNames.OUTPUT_DATA_SCHEMA: output_data_schema
                                }
            published_model = self.client.repository.store_model(model=self.model_path, meta_props=model_meta_props)
            TestWMLClientWithDO.model_uid = self.client.repository.get_model_uid(published_model)
            TestWMLClientWithDO.logger.info("Published model ID:" + str(TestWMLClientWithDO.model_uid))
            self.assertIsNotNone(TestWMLClientWithDO.model_uid)

    def test_04_get_details(self):
        if not TestWMLClientWithDO.skip_tests:
            TestWMLClientWithDO.logger.info("Get details")
            self.assertIsNotNone(self.client.repository.get_details(TestWMLClientWithDO.model_uid))

    def test_05_get_model_details(self):
        if not TestWMLClientWithDO.skip_tests:
            TestWMLClientWithDO.logger.info("Get model details")
            self.assertIsNotNone(self.client.repository.get_model_details(TestWMLClientWithDO.model_uid))

    def test_07_create_deployment(self):
        if not TestWMLClientWithDO.skip_tests:
            TestWMLClientWithDO.logger.info("Create deployment")

            deploy_meta = {
                self.client.deployments.ConfigurationMetaNames.NAME: "deployment_DO",
                self.client.deployments.ConfigurationMetaNames.DESCRIPTION: "DO deployment",
                self.client.deployments.ConfigurationMetaNames.BATCH: {},
                self.client.deployments.ConfigurationMetaNames.HARDWARE_SPEC: {"name": "S", "num_nodes": 1}
            }


            deployment_details = self.client.deployments.create(TestWMLClientWithDO.model_uid,
                                                                meta_props=deploy_meta)

            TestWMLClientWithDO.logger.debug("Deployment details: {}".format(deployment_details))

            TestWMLClientWithDO.deployment_uid = self.client.deployments.get_uid(deployment_details)
            TestWMLClientWithDO.logger.debug("Deployment uid: {}".format(TestWMLClientWithDO.deployment_uid))


    def test_08_get_deployment_details(self):
        if not TestWMLClientWithDO.skip_tests:
            TestWMLClientWithDO.logger.info("Get deployment details")
            deployment_details = self.client.deployments.get_details(deployment_uid=TestWMLClientWithDO.deployment_uid)
            self.assertIsNotNone(deployment_details)


    def test_09_create_job(self):
        if not TestWMLClientWithDO.skip_tests:
            TestWMLClientWithDO.logger.info("Create job details")
            TestWMLClientWithDO.logger.debug("Create job")

            # initialize list of lists
            diet_food = pd.DataFrame([["Roasted Chicken", 0.84, 0, 10],
                                      ["Spaghetti W/ Sauce", 0.78, 0, 10],
                                      ["Tomato,Red,Ripe,Raw", 0.27, 0, 10],
                                      ["Apple,Raw,W/Skin", 0.24, 0, 10],
                                      ["Grapes", 0.32, 0, 10],
                                      ["Chocolate Chip Cookies", 0.03, 0, 10],
                                      ["Lowfat Milk", 0.23, 0, 10],
                                      ["Raisin Brn", 0.34, 0, 10],
                                      ["Hotdog", 0.31, 0, 10]], columns=["name", "unit_cost", "qmin", "qmax"])

            diet_food_nutrients = pd.DataFrame([
                ["Spaghetti W/ Sauce", 358.2, 80.2, 2.3, 3055.2, 11.6, 58.3, 8.2],
                ["Roasted Chicken", 277.4, 21.9, 1.8, 77.4, 0, 0, 42.2],
                ["Tomato,Red,Ripe,Raw", 25.8, 6.2, 0.6, 766.3, 1.4, 5.7, 1],
                ["Apple,Raw,W/Skin", 81.4, 9.7, 0.2, 73.1, 3.7, 21, 0.3],
                ["Grapes", 15.1, 3.4, 0.1, 24, 0.2, 4.1, 0.2],
                ["Chocolate Chip Cookies", 78.1, 6.2, 0.4, 101.8, 0, 9.3, 0.9],
                ["Lowfat Milk", 121.2, 296.7, 0.1, 500.2, 0, 11.7, 8.1],
                ["Raisin Brn", 115.1, 12.9, 16.8, 1250.2, 4, 27.9, 4],
                ["Hotdog", 242.1, 23.5, 2.3, 0, 0, 18, 10.4]
            ], columns=["Food", "Calories", "Calcium", "Iron", "Vit_A", "Dietary_Fiber", "Carbohydrates", "Protein"])

            diet_nutrients = pd.DataFrame([
                ["Calories", 2000, 2500],
                ["Calcium", 800, 1600],
                ["Iron", 10, 30],
                ["Vit_A", 5000, 50000],
                ["Dietary_Fiber", 25, 100],
                ["Carbohydrates", 0, 300],
                ["Protein", 50, 100]
            ], columns=["name", "qmin", "qmax"])

            job_payload_ref = {
                self.client.deployments.DecisionOptimizationMetaNames.INPUT_DATA: [
                    {
                        "id": "diet_food.csv",
                        "values": diet_food
                    },
                    {
                        "id": "diet_food_nutrients.csv",
                        "values": diet_food_nutrients
                    },
                    {
                        "id": "diet_nutrients.csv",
                        "values": diet_nutrients
                    }
                ],
                self.client.deployments.DecisionOptimizationMetaNames.OUTPUT_DATA: [
                    {
                        "id": ".*.csv"
                    }
                ]
            }

            TestWMLClientWithDO.job_details = self.client.deployments.create_job(TestWMLClientWithDO.deployment_uid,
                                                                                 meta_props=job_payload_ref)
            TestWMLClientWithDO.job_id = self.client.deployments.get_job_uid(TestWMLClientWithDO.job_details)


    def test_10_get_job_status(self):
        if not TestWMLClientWithDO.skip_tests:
            start_time = time.time()
            diff_time = start_time - start_time
            while True and diff_time < 10 * 60:
                time.sleep(3)
                response = self.client.deployments.get_job_status(TestWMLClientWithDO.job_id)
                if response['state'] == 'completed' or response['state'] == 'error' or response['state'] == 'failed':
                    break
                diff_time = time.time() - start_time

            self.assertIsNotNone(response)
            self.assertTrue(response['state'] == 'completed')

    def test_11_extract_and_display_solution(self):
        if not TestWMLClientWithDO.skip_tests:
            job_details_do = self.client.deployments.get_job_details(TestWMLClientWithDO.job_id)
            kpi = job_details_do['entity']['decision_optimization']['solve_state']['details']['KPI.Total Calories']
            print(kpi)
            TestWMLClientWithDO.logger.debug("The value of kpi is ", kpi)
            self.assertTrue(kpi == '2000.0')


    def test_12_list_jobs(self):
        if not TestWMLClientWithDO.skip_tests:
            self.client.deployments.list_jobs()

    def test_12_delete_job(self):
        if not TestWMLClientWithDO.skip_tests:
            self.client.deployments.delete_job(TestWMLClientWithDO.job_id)

    def test_13_delete_deployment(self):
        if not TestWMLClientWithDO.skip_tests:
            TestWMLClientWithDO.logger.info("Delete deployment")
            self.client.deployments.delete(TestWMLClientWithDO.deployment_uid)

    def test_14_delete_model(self):
        if not TestWMLClientWithDO.skip_tests:
            TestWMLClientWithDO.logger.info("Delete function")
            self.client.script.delete(TestWMLClientWithDO.model_uid)

    def test_15_delete_space(self):
        if not TestWMLClientWithDO.skip_tests:
            self.client.spaces.delete(TestWMLClientWithDO.space_id)


if __name__ == '__main__':
    unittest.main()



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import inspect
import sys
import getpass
import os
from py4j.java_collections import MapConverter,ListConverter
from pyspark import SparkContext
from pyspark.ml.pipeline import Pipeline
from pyspark.ml.util import keyword_only
from pyspark.sql.dataframe import DataFrame

from DAG import DAG
from Result import DataResult,ModelResult,SinkResult
from IBMSparkPipelineModel import IBMSparkPipelineModel
from Wrapper import PythonJavaConversions, JavaParams

class IBMSparkPipeline (Pipeline):
    """
    Base class for creating ML and ETL Pipeline. This class extends from Spark ML Pipeline.
    """
    _sc = SparkContext._active_spark_context
    _logSrcLang = "Py:"

    @keyword_only
    def __init__(self, stages=None):

        self._jLogger = self._sc._jvm.org.apache.log4j.Logger
        self.logger = self._jLogger.getLogger(self.__class__.__name__)
        methodname = str(inspect.stack()[0][3])
        logMsg = self._logSrcLang + self.__class__.__name__ + ":" + methodname + ": [Params: " + "stages => " + str(
            stages) + "]"
        self.logger.info(logMsg)

        super(IBMSparkPipeline, self).__init__()

        self._jPipeline = self._sc._jvm.com.ibm.analytics.ngp.pipeline.pythonbinding.Pipelines
        self.jIBMSparkPipeline = self._jPipeline.getIBMSparkPipeline().apply()
        self._to_map = self._jPipeline.toScalaMap

        self.jTkObj = self._sc._jvm.com.ibm.analytics.ngp.util.NotebookRequestParams
        self.blueId = getpass.getpass('ibm.ax.blueid')
        self.orgId = getpass.getpass('ibm.ax.orgid')
        self.jTkObj.setBlueIdToken(self.blueId)
        self.jTkObj.setOrgId(self.orgId)

        self.tenantId = os.environ['SPARK_TENANT_ID']
        self.jTkObj.setTenant(self.tenantId)

    def __setIBMSparkPipelineObject(self, IBMSparkPipelineObj):
        self.jIBMSparkPipeline = IBMSparkPipelineObj

    def save(self, path):
        """
        Saves an :class:`IBMSparkPipeline` object to GIT.

        :param filename: filename to be saved in GIT.
        :return: None.
        """
        methodname = str(inspect.stack()[0][3])
        logMsg = self._logSrcLang + self.__class__.__name__ + ":" + methodname + ": [Params: " + " filename => " + str(
            path) + "]"
        self.logger.info(logMsg)

        if not isinstance(path, basestring):
            logMsg = self._logSrcLang + self.__class__.__name__ + ":" + methodname + ": path should be a " \
                    "basestring, got " + str(type(path))
            self.logger.error(logMsg)
            raise TypeError("path should be a basestring, got %s" % type(path))

        if self.getStages() is not None:
            self._to_java().save(path)
        else:
            self.jIBMSparkPipeline.save(path)


    def saveGpfs (self, path):
        """
        Saves an :class:`IBMSparkPipeline` object to GPFS.

        :param filename: filename to be saved in GPFS.
        :return: None
        """
        methodname = str(inspect.stack()[0][3])
        logMsg = self._logSrcLang + self.__class__.__name__ + ":" + methodname + ": [Params: " + " filename => " + str(
                path) + "]"
        self.logger.info(logMsg)

        if self.getStages() is not None:
            self._to_java().writer().saveGpfs(path)
        else:
            self.jIBMSparkPipeline.writer().saveGpfs(path)


    def setStages(self, stages):
        """
        Sets the stages of an ML or ETL Pipeline.

        :param stages: A list of Pipeline stages for a ML Pipeline; a :class:`pipeline.DAG.DAG`
                      object for an ETL Pipeline.
        :return: An :class:`IBMSparkPipeline` object with stages set.
        """
        methodname = str(inspect.stack()[0][3])
        logMsg = self._logSrcLang + self.__class__.__name__ + ":" + methodname + ": [Params: " \
                 + " stages => " + str(stages) + "]"
        self.logger.info(logMsg)

        if type(stages) is not list and not isinstance(stages,DAG):
            logMsg = self._logSrcLang + self.__class__.__name__ + ":" + methodname + ": Input for setStages()" \
                    "should be a list of stages for ML Pipeline,  " \
                    "DAG for ETL Pipeline, got " + str(type(stages))
            self.logger.error(logMsg)
            raise TypeError("Input for setStages() should be a list of stages for ML Pipeline, "
                            "DAG for ETL Pipeline, got %s" % type(stages))

        if isinstance(stages, DAG):
            self.jIBMSparkPipeline = self.jIBMSparkPipeline.setStages(stages._to_java())
        else:
            super(IBMSparkPipeline, self).setStages(stages)
        return self

    def _to_java_stages(self):
        methodname = str(inspect.stack()[0][3])
        logMsg = self._logSrcLang + self.__class__.__name__ + ":" + methodname + ": [Params: None]"
        self.logger.info(logMsg)

        gateway = SparkContext._gateway
        cls = SparkContext._jvm.org.apache.spark.ml.PipelineStage
        java_stages = gateway.new_array(cls, len(self.getStages()))
        for idx, stage in enumerate(self.getStages()):
            java_stages[idx] = PythonJavaConversions._to_java_stage(stage)

        return java_stages

    def fit(self, dataset=None):
        """
        Runs an ML or ETL Pipeline

        :param dataset: A DataFrame for a ML Pipeline; None for ETL Pipeline.
        :return: An :class:`pipeline.IBMSparkPipelineModel.IBMSparkPipelineModel` object for ML Pipeline.

                An array of :class:`pipeline.Result.Result` for ETL Pipeline.

                The Result can be either of:

                - :class:`pipeline.Result.DataResult` : If the end stage of a DAG is a Transformer.
                - :class:`pipeline.Result.ModelResult` : If the end stage of a DAG is an Estimator.
                - :class:`pipeline.Result.SinkResult`: If the end stage of a DAG is :class:`pipeline.Sink.Sink`
        """
        methodname = str(inspect.stack()[0][3])
        logMsg = self._logSrcLang + self.__class__.__name__ + ":" + methodname + ": [Params: " + " dataset => " + str(
                dataset) + "]"
        self.logger.info(logMsg)

        if dataset is not None and not isinstance(dataset,DataFrame):
            logMsg = self._logSrcLang + self.__class__.__name__ + ":" + methodname + "Input dataset for fit() " \
                    "should be a DataFrame, got " + str(type(dataset))
            self.logger.error(logMsg)
            raise TypeError("Input dataset for fit() should be a DataFrame, got %s" % type(dataset))

        if self.getStages() is not None:
            if dataset is not None:
                jIBMSparkPipelineModel = self.jIBMSparkPipeline.setStages(self._to_java_stages()).fit(dataset._jdf)
                return IBMSparkPipelineModel._from_java(jIBMSparkPipelineModel)
            else:
                self.jIBMSparkPipeline.fit()

        else:
            if dataset is None:
                sca_seq = self.jIBMSparkPipeline.fit()
                jlist = PythonJavaConversions._to_java_list(self._jPipeline, sca_seq)
                fullclassname = sca_seq.head().getClass().toString()
                classname = fullclassname.split(".")[-1]
                resultclass = getattr(sys.modules[__name__], classname)
                resultlist = resultclass._populateResult(jlist)
                return resultlist
            else:
                self.jIBMSparkPipeline.fit(dataset._jdf)

    def updateBindings (self, binddict):
        """
        Updates the input DataFrame to be passed to the DAG root node(s) after a saved ETL Pipeline
        is loaded back.

        :param binddict: A dictionary that specifies the label and input DataFrame
                         for the DAG root node.
        :return: An :class:`IBMSparkPipeline` object with bindings updated.
        """
        methodname = str(inspect.stack()[0][3])
        logMsg = self._logSrcLang + self.__class__.__name__ + ":" + methodname + ": [Params: " + " binddict => " + str(
            binddict) + "]"
        self.logger.info(logMsg)

        if type(binddict) is not dict :
            logMsg = self._logSrcLang + self.__class__.__name__ + ":" + methodname + ": Label and Input DataFrame for" \
                    "updateBindings() must be specified as a dictionary, got " + str(type(binddict))
            self.logger.error(logMsg)
            raise TypeError ("Label and Input DataFrame for updateBindings() must be specified as a  dictionary, "
                             "got %s" % type(binddict))

        partialJbinds = {}
        for k, v in binddict.iteritems():
            partialJbinds[k] = v._jdf

        jbinds = MapConverter().convert(partialJbinds, self._sc._gateway._gateway_client)
        sca_map = self._to_map(jbinds)
        self.jIBMSparkPipeline = self.jIBMSparkPipeline.updateBindings(sca_map)
        return self

    def getRootLabels(self) :
        """
        Returns the Root Labels for the DAG root node(s).
        """
        slabellist = self.jIBMSparkPipeline.getRootLabels()
        jlabellist = PythonJavaConversions._to_java_list(self._jPipeline, slabellist)
        plabellist=[]
        for index in range(len(jlabellist)):
            plabellist.append(str(jlabellist[index]))
        return plabellist


    @classmethod
    def load(cls, path):
        """
        Loads an :class:`IBMSparkPipeline` object from the saved path in GPFS.

        :param filename: filename which was saved in GPFS
        :return: A :class:`IBMSparkPipeline` object.
        """
        logger = cls._sc._jvm.org.apache.log4j.Logger.getLogger(cls.__name__)
        methodname = str(inspect.stack()[0][3])
        logMsg = cls._logSrcLang + cls.__name__ + ":" + methodname + ": [Params: " + " path => " + str(path) + "]"
        logger.info(logMsg)

        if not isinstance(path, basestring):
            logMsg = cls._logSrcLang + cls.__name__ + ":" + methodname + ": path should be a basestring, got " \
                     + str(type(path))
            cls.logger.error(logMsg)
            raise TypeError("path should be a basestring, got type %s" % type(path))
        else:
            _jread = PythonJavaConversions._load_java_obj(cls)
            java_obj = _jread.load(path)

        numstages = java_obj.getNonLinearStages().getStages().length()
        # If the below check returns 0, then its an ML Pipeline object. Otherwise its an ETL Pipleline object.
        if numstages == 0:
            return cls._from_java(java_obj)
        else:
            loadpipeline = IBMSparkPipeline()
            loadpipeline.__setIBMSparkPipelineObject(java_obj)
            return loadpipeline

    @classmethod
    def loadGpfs(cls, path):
        """
        Loads an :class:`IBMSparkPipeline` object from the saved path in GIT.

        :param filename: filename which was saved in GIT.
        :return: An :class:`IBMSparkPipeline` object.
        """
        logger = cls._sc._jvm.org.apache.log4j.Logger.getLogger(cls.__name__)
        methodname = str(inspect.stack()[0][3])
        logMsg = cls._logSrcLang + cls.__name__ + ":" + methodname + ": [Params: " + " path => " + str(path) + "]"
        logger.info(logMsg)

        # return Reader(cls).load(path)

        _jread = PythonJavaConversions._load_java_obj(cls)

        if not isinstance(path, basestring):
            raise TypeError("path should be a basestring, got type %s" % type(path))

        else:
            java_obj = _jread.reader().loadGpfs(path)

        numstages = java_obj.getNonLinearStages().getStages().length()
        # If the below check returns 0, then its an ML Pipeline object. Otherwise its an ETL Pipleline object.
        if numstages == 0:
            return cls._from_java(java_obj)
        else:
            loadpipeline = IBMSparkPipeline()
            loadpipeline.__setIBMSparkPipelineObject(java_obj)
            return loadpipeline

    @classmethod
    def _from_java(cls, java_stage):
        """
        Given a Java Pipeline, create and return a Python wrapper of it.
        Used for ML persistence.
        """
        # Create a new instance of this stage.
        logger = cls._sc._jvm.org.apache.log4j.Logger.getLogger(cls.__name__)
        methodname = str(inspect.stack()[0][3])
        logMsg = cls._logSrcLang + cls.__name__ + ":" + methodname + ": [Params: " + " java_stage => " + str(
            java_stage) + "]"
        logger.info(logMsg)

        py_stage = cls()
        # Load information from java_stage to the instance.
        py_stages = [PythonJavaConversions._from_java_stage(s) for s in java_stage.getStages()]
        py_stage.setStages(py_stages)
        py_stage = PythonJavaConversions._resetUid(py_stage, java_stage.uid())
        return py_stage

    def _to_java(self):
        """
        Transfer this instance to a Java Pipeline.  Used for ML persistence.

        :return: Java object equivalent to this instance.
        """
        self.logger = self._jLogger.getLogger(self.__class__.__name__)
        methodname = str(inspect.stack()[0][3])
        logMsg = self._logSrcLang + self.__class__.__name__ + ":" + methodname + ": [Params: None]"
        self.logger.info(logMsg)

        gateway = SparkContext._gateway
        cls = SparkContext._jvm.org.apache.spark.ml.PipelineStage
        java_stages = gateway.new_array(cls, len(self.getStages()))
        for idx, stage in enumerate(self.getStages()):
            java_stages[idx] = PythonJavaConversions._to_java_stage(stage)

        _java_obj = JavaParams._new_java_obj("com.ibm.analytics.ngp.pipeline.IBMSparkPipeline",
                                             self.uid)
        _java_obj.setStages(java_stages)

        return _java_obj





#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest

import ibm_boto3

from ibm_watson_machine_learning.helpers.connections import DataConnection, S3Location, S3Connection
from ibm_watson_machine_learning.utils.autoai.errors import CannotReadSavedRemoteDataBeforeFit, WMLClientError
from ibm_watson_machine_learning.tests.utils import bucket_exists, create_bucket, is_cp4d
from ibm_watson_machine_learning.tests.autoai.abstract_tests_classes import (
    AbstractTestAutoAIRemote)


@unittest.skip("S3 type no longer supported")
class TestAutoAIRemote(AbstractTestAutoAIRemote, unittest.TestCase):
    """
    The test can be run on CLOUD
    The test covers:
    - COS connection set-up
    - Saving data `iris.csv` to data assets
    - downloading training data from data assets
    - downloading all generated pipelines to lale pipeline
    - deployment with lale pipeline
    - deployment deletion
    Connection used in test:
     - input: ConnectedDataAsset pointing to COS.
     - output: S3 connection pointing to COS.
    """

    def test_00b_prepare_COS_instance(self):
        TestAutoAIRemote.cos_resource = ibm_boto3.resource(
            service_name="s3",
            endpoint_url=self.cos_endpoint,
            aws_access_key_id=self.cos_credentials['cos_hmac_keys']['access_key_id'],
            aws_secret_access_key=self.cos_credentials['cos_hmac_keys']['secret_access_key']
        )
        # Prepare bucket
        if not bucket_exists(self.cos_resource, self.bucket_name):
            TestAutoAIRemote.bucket_name = create_bucket(self.cos_resource, self.bucket_name)


            self.assertIsNotNone(TestAutoAIRemote.bucket_name)
            self.assertTrue(bucket_exists(self.cos_resource, TestAutoAIRemote.bucket_name))

        self.cos_resource.Bucket(self.bucket_name).upload_file(
            self.data_location,
            self.data_cos_path
        )

    def test_00c_prepare_connection_to_COS(self):
        connection_details = self.wml_client.connections.create({
            'datasource_type': self.wml_client.connections.get_datasource_type_uid_by_name('bluemixcloudobjectstorage'),
            'name': 'Connection to COS for tests',
            'properties': {
                'bucket': self.bucket_name,
                'access_key': self.cos_credentials['cos_hmac_keys']['access_key_id'],
                'secret_key': self.cos_credentials['cos_hmac_keys']['secret_access_key'],
                'iam_url': self.wml_client.service_instance._href_definitions.get_iam_token_url(),
                'url': self.cos_endpoint
            }
        })

        TestAutoAIRemote.connection_id = self.wml_client.connections.get_uid(connection_details)
        self.assertIsInstance(self.connection_id, str)

    def test_02_DataConnection_setup(self):
        TestAutoAIRemote.data_connection = DataConnection(
            connection_asset_id=self.connection_id,
            location=S3Location(
                bucket=self.bucket_name,
                path=self.data_cos_path
            )
        )
        TestAutoAIRemote.results_connection = DataConnection(
            connection=S3Connection(
                endpoint_url=self.cos_endpoint,
                access_key_id=self.cos_credentials['cos_hmac_keys']['access_key_id'],
                secret_access_key=self.cos_credentials['cos_hmac_keys']['secret_access_key']
            ),
            location=S3Location(
                bucket=self.bucket_name,
                path=self.results_cos_path
            )
        )

        self.assertIsNotNone(obj=TestAutoAIRemote.data_connection)
        self.assertIsNotNone(obj=TestAutoAIRemote.results_connection)

    def test_02a_read_saved_remote_data_before_fit(self):
        self.assertRaises(NotImplementedError, self.data_connection.read)

    def test_29_delete_connection_and_connected_data_asset(self):
        self.wml_client.connections.delete(self.connection_id)

        with self.assertRaises(WMLClientError):
            self.wml_client.connections.get_details(self.connection_id)


if __name__ == '__main__':
    unittest.main()



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from .progress_bar import ProgressBar
from .utils import is_ipython

__all__ = [
    "LocalTrainingMessageHandler"
]


class LocalTrainingMessageHandler:
    """Show progress bar during local AutoPipelines training/fit."""
    def __init__(self):
        self.progress_bar = None
        self.progress_bar_1 = None
        self.previous_stage = None

        # note: only notebook version
        if is_ipython() and False:
            self.total = 100
            self.progress_bar_1 = ProgressBar(desc="Total", total=self.total, position=0, ncols='100%')
            self.progress_bar_2 = ProgressBar(desc="Waiting", total=self.total, leave=False, ncols='100%')

        # note: only console version
        else:
            self.total = 200
            self.progress_bar = ProgressBar(desc="Total", total=self.total, position=0, ncols=100)

    def on_training_message(self, status_state):
        """This method should be used by the ai4ml estimator/optimization process during training."""

        if 'ml_metrics' in status_state:
            metric = status_state['ml_metrics']
            for stage in ['pre_hpo_d_output', 'hpo_d_output', 'hpo_c_output', 'fold_output',
                          'cognito_output', 'compose_model_type_output', 'global_output', 'daub_running_output',
                          'cognito_running_output', 'hpo_c_running_output', 'hpo_d_running_output']:

                if (stage in metric) and (metric[stage] is not None):

                    if self.progress_bar is not None:
                        self.progress_bar.set_description(desc=stage)
                        if self.total - self.progress_bar.counter <= 5:
                            pass

                        else:
                            self.progress_bar.increment_counter(progress=1)
                            self.progress_bar.update()

                    else:

                        if self.previous_stage != stage:
                            self.previous_stage = stage
                            self.progress_bar_2.last_update()
                            self.progress_bar_2.close()
                            self.progress_bar_2 = ProgressBar(desc="Waiting", total=self.total, leave=False, ncols='100%')
                            if self.total - self.progress_bar_1.counter <= 10:
                                pass

                            else:
                                self.progress_bar_1.increment_counter(progress=5)
                                self.progress_bar_1.update()

                        self.progress_bar_2.set_description(desc=stage)
                        if self.total - self.progress_bar_1.counter <= 1:
                            pass

                        else:
                            self.progress_bar_2.increment_counter(progress=1)
                            self.progress_bar_2.update()



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import pathlib
from environs import Env

env = Env()
env.read_env()

# export FL_WORKING_DIR=/home/username/
# export FL_MODEL_DIR=/home/username/model


def create_dir(path):
    pathlib.Path(path).mkdir(parents=True, exist_ok=True)


with env.prefixed("FL_"):
    working_directory = env("WORKING_DIR", os.getcwd())
    create_dir(working_directory)

    model_directory = env("MODEL_DIR", working_directory)
    create_dir(model_directory)
